from pathlib import Path
import shutil
import warnings

import numpy as np
import pandas as pd

import openmdao.api as om

import wisdem.orbit.orbit_api as orbit_wisdem

from ORBIT.core.library import default_library
from ORBIT.core.library import initialize_library

from ard.cost.wisdem_wrap import ORBIT_setup_latents


def generate_orbit_location_from_graph(
    graph,  # TODO: replace with a terse_links representation
    X_turbines,
    Y_turbines,
    X_substations,
    Y_substations,
    allow_branching_approximation=False,
):
    """
    go from a optiwindnet graph to an ORBIT input CSV

    convert a optiwindnet graph representation of a collection system and get to
    a best-possible approximation of the same collection system for
    compatibility with ORBIT. ORBIT doesn't allow branching and optiwindnet does
    by default, so we allow some cable duplication if necessary to get a
    conservative approximation of the BOS costs if the graph isn't compatible
    with ORBIT

    Parameters
    ----------
    graph : networkx.Graph
        the graph representation of the collection system design
    X_turbines : np.array
        the cartesian X locations, in kilometers, of the turbines
    Y_turbines : np.array
        the cartesian Y locations, in kilometers, of the turbines
    X_substations : np.array
        the cartesian X locations, in kilometers, of the substations
    Y_substations : np.array
        the cartesian Y locations, in kilometers, of the substations

    Returns
    -------
    pandas.DataFrame
        a dataframe formatted for ORBIT to specify a farm layout

    Raises
    ------
    RecursionError
        if the recursive setup seems to be stuck in a loop
    """

    # get all edges, sorted by the first node then the second node
    edges_to_process = [edge for edge in graph.edges]
    edges_to_process.sort(key=lambda x: (x[0], x[1]))
    # get the edges with a negative index node (a substation)
    edges_inclsub = [edge for edge in edges_to_process if edge[0] < 0 or edge[1] < 0]
    edges_inclsub.sort(key=lambda x: (x[0], x[1]))

    # check to see if any nodes appear more than twice
    # (i.e. once destination and possibly one source)
    node_countmap = dict.fromkeys(
        list(set(node for edge in edges_to_process for node in edge)), 0
    )
    for edge in edges_to_process:
        node_countmap[edge[0]] += 1
        node_countmap[edge[1]] += 1
    # if this has branching, handle it
    if np.any(
        (
            np.array(list(node_countmap.values())) > 2
        )  # multiple turbine appearances indicates a branch
        & (
            np.array(list(node_countmap.keys())) >= 0
        )  # but substations do appear so "mask" them
    ):
        if allow_branching_approximation:
            warnings.warn(
                "The provided collection system design graph includes branching, "
                "which ORBIT does not support. Proceeding with an approximate "
                "radial collection system for cost modeling."
            )
        else:
            raise ValueError(
                "The graph has branching. ORBIT does not support this. "
                "By modifying the approximate_branches option to True in the "
                "ORBITDetail component, you can allow ORBIT to approximate the "
                "BOS costs by a close radial-layout collection system "
                "approximation."
            )

    # data for ORBIT
    data_orbit = {
        "id": [],
        "substation_id": [],
        "name": [],
        "longitude": [],
        "latitude": [],
        "string": [],
        "order": [],
        "cable_length": [],
        "bury_speed": [],
    }

    idx_string = 0
    order = 0

    for edge in edges_inclsub:  # every edge w/ a substation starts a string

        def handle_edge(
            edge, turbine_origination, idx_string, order, recursion_level=0
        ):
            # recursively handle the edges

            if recursion_level > 10:  # for safe recursion
                raise RecursionError("Recursion limit reached")

            # get the target turbine index
            turbine_tgt_index = edge[0] if edge[0] != turbine_origination else edge[1]
            # get the turbine name
            turbine_name = turbine_id = f"t{turbine_tgt_index:03d}"

            # add the turbine to the dataset
            data_orbit["id"].append(turbine_id)
            data_orbit["substation_id"].append(substation_id)
            data_orbit["name"].append(turbine_name)
            data_orbit["longitude"].append(X_turbines[turbine_tgt_index])
            data_orbit["latitude"].append(Y_turbines[turbine_tgt_index])
            data_orbit["string"].append(int(idx_string))
            data_orbit["order"].append(int(order))
            data_orbit["cable_length"].append(0)  # ORBIT computes automatically
            data_orbit["bury_speed"].append(0)  # ORBIT computes automatically

            # pop this edge out of the edges list
            edges_to_process.remove(edge)

            # get the set of remaining edges that include the terminal turbine
            edges_turbine = [e for e in edges_to_process if (turbine_tgt_index in e)]

            order += 1

            for new_string, edge_next in enumerate(edges_turbine):
                if new_string:
                    idx_string += 1
                    order = 0
                idx_string, order = handle_edge(
                    edge_next,
                    turbine_tgt_index,
                    idx_string,
                    order,
                    recursion_level=recursion_level + 1,
                )

            return idx_string, order

        # get the substation id as a one-liner
        substation_index = len(X_substations) + (edge[0] if edge[0] < 0 else edge[1])
        # get the substation name
        substation_name = substation_id = f"oss{substation_index:01d}"

        # add the substation to the dataset
        if not substation_id in data_orbit["id"]:
            data_orbit["id"].append(substation_id)
            data_orbit["substation_id"].append(substation_id)
            data_orbit["name"].append(substation_name)
            data_orbit["longitude"].append(X_substations[substation_index] / 1.0e3)
            data_orbit["latitude"].append(Y_substations[substation_index] / 1.0e3)
            data_orbit["string"].append(None)
            data_orbit["order"].append(None)
            data_orbit["cable_length"].append(None)
            data_orbit["bury_speed"].append(None)

        # handle the edge that we get
        idx_string, order = handle_edge(
            edge, substation_index - len(X_substations), idx_string, order
        )

        order = 0
        idx_string += 1

    df_orbit = pd.DataFrame(data_orbit).fillna("")
    df_orbit.string = [int(v) if v != "" else "" for v in df_orbit.string]
    df_orbit.order = [int(v) if v != "" else "" for v in df_orbit.order]

    return df_orbit


class ORBITDetail(orbit_wisdem.Orbit):
    """
    Wrapper for WISDEM's ORBIT offshore BOS calculators.

    A thicker wrapper of `wisdem.orbit_api` that 1) replaces capabilities that
    assume a grid farm layout that is default in WISDEM's ORBIT with a custom
    array layout, and 2) traps warning messages that are recognized not to be
    issues.

    See: https://github.com/NLRWindSystems/ORBIT
    """

    def initialize(self):
        """Initialize for API connections."""
        super().initialize()

        self.options.declare("case_title", default="working")
        self.options.declare("modeling_options")
        self.options.declare("approximate_branches", default=False)

    def setup(self):
        """Define all input variables from all models."""

        # load modeling options
        self.modeling_options = self.options["modeling_options"]
        self.N_turbines = self.modeling_options["layout"]["N_turbines"]
        self.N_substations = self.modeling_options["layout"]["N_substations"]

        self.set_input_defaults("wtiv", "example_wtiv")
        self.set_input_defaults("feeder", "example_feeder")
        # self.set_input_defaults("num_feeders", 1)
        # self.set_input_defaults("num_towing", 1)
        # self.set_input_defaults("num_station_keeping", 3)
        # self.set_input_defaults(
        #    "oss_install_vessel", "example_heavy_lift_vessel",
        # )
        self.set_input_defaults("site_distance", 40.0, units="km")
        self.set_input_defaults("site_distance_to_landfall", 40.0, units="km")
        self.set_input_defaults("interconnection_distance", 40.0, units="km")
        self.set_input_defaults("plant_turbine_spacing", 7)
        self.set_input_defaults("plant_row_spacing", 7)
        self.set_input_defaults("plant_substation_distance", 1, units="km")
        # self.set_input_defaults("num_port_cranes", 1)
        # self.set_input_defaults("num_assembly_lines", 1)
        self.set_input_defaults("takt_time", 170.0, units="h")
        self.set_input_defaults("port_cost_per_month", 2e6, units="USD/mo")
        self.set_input_defaults("construction_insurance", 44.0, units="USD/kW")
        self.set_input_defaults("construction_financing", 183.0, units="USD/kW")
        self.set_input_defaults("contingency", 316.0, units="USD/kW")
        self.set_input_defaults("commissioning_cost_kW", 44.0, units="USD/kW")
        self.set_input_defaults("decommissioning_cost_kW", 58.0, units="USD/kW")
        self.set_input_defaults("site_auction_price", 100e6, units="USD")
        self.set_input_defaults("site_assessment_cost", 50e6, units="USD")
        self.set_input_defaults("construction_plan_cost", 1e6, units="USD")
        self.set_input_defaults("installation_plan_cost", 2.5e5, units="USD")
        self.set_input_defaults("boem_review_cost", 0.0, units="USD")

        self.add_subsystem(
            "orbit",
            ORBITWisdemDetail(
                modeling_options=self.modeling_options,
                case_title=self.options["case_title"],
                approximate_branches=self.options["approximate_branches"],
                floating=self.modeling_options["floating"],
                jacket=self.modeling_options.get("jacket"),
                jacket_legs=self.modeling_options.get("jacket_legs"),
            ),
            promotes=["*"],
        )


class ORBITWisdemDetail(orbit_wisdem.OrbitWisdem):
    """ORBIT-WISDEM Fixed Substructure API, modified for detailed layouts"""

    _path_library = None

    def initialize(self):
        super().initialize()

        self.options.declare("case_title", default="working")
        self.options.declare("modeling_options")
        self.options.declare("approximate_branches", default=False)

    def setup(self):
        """Define all the inputs."""

        # call the superclass method
        super().setup()

        # load modeling options
        self.modeling_options = self.options["modeling_options"]
        self.N_turbines = self.modeling_options["layout"]["N_turbines"]
        self.N_substations = self.modeling_options["layout"]["N_substations"]

        # bring in collection system design
        self.add_discrete_input("graph", None)

        # add the detailed turbine and substation locations
        self.add_input("x_turbines", np.zeros((self.N_turbines,)), units="km")
        self.add_input("y_turbines", np.zeros((self.N_turbines,)), units="km")
        self.add_input("x_substations", np.zeros((self.N_substations,)), units="km")
        self.add_input("y_substations", np.zeros((self.N_substations,)), units="km")

        # copy the default ORBIT library to a local directory under case_files
        path_library_default = Path(default_library).absolute()
        self._path_library = (
            Path("case_files") / self.options["case_title"] / "ORBIT_library"
        ).absolute()
        if path_library_default.exists():
            shutil.copytree(
                path_library_default, self._path_library, dirs_exist_ok=True
            )
        else:
            raise FileNotFoundError(
                f"Can not find default ORBIT library at {path_library_default}."
            )

    def compile_orbit_config_file(
        self,
        inputs,
        outputs,
        discrete_inputs,
        discrete_outputs,
    ):

        config = super().compile_orbit_config_file(
            inputs,
            outputs,
            discrete_inputs,
            discrete_outputs,
        )  # run the superclass

        # remove the grid plant option, and replace with a custom plant
        config["plant"] = {
            "layout": "custom",
            "num_turbines": int(discrete_inputs["number_of_turbines"]),
            "turbine_spacing": inputs["plant_turbine_spacing"],
            "row_spacing": inputs["plant_row_spacing"],
        }

        # switch to the custom array system design
        if not ("ArraySystemDesign" in config["design_phases"]):
            raise KeyError(
                "I assumed that 'ArraySystemDesign' would be in the config. Something changed."
            )
        config["design_phases"][
            config["design_phases"].index("ArraySystemDesign")
        ] = "CustomArraySystemDesign"

        # add a turbine location csv on the config
        basename_farm_location = "wisdem_detailed_array"
        config["array_system_design"]["distance"] = True  # don't use WGS84 lat/long
        config["array_system_design"]["location_data"] = basename_farm_location
        config["array_system_design"]["cables"] = [
            f"XLPE_185mm_66kV{'_dynamic' if self.options['floating'] else ''}",
            f"XLPE_500mm_132kV{'_dynamic' if self.options['floating'] else ''}",
            f"XLPE_630mm_66kV{'_dynamic' if self.options['floating'] else ''}",
            f"XLPE_1000mm_220kV{'_dynamic' if self.options['floating'] else ''}",
        ]  # we require bigger cables than the standard WISDEM wrap

        # create the csv file that holds the farm layout
        path_farm_location = (
            self._path_library / "cables" / (basename_farm_location + ".csv")
        )

        # generate the csv data needed to locate the farm elements
        generate_orbit_location_from_graph(
            discrete_inputs["graph"],
            inputs["x_turbines"],
            inputs["y_turbines"],
            inputs["x_substations"],
            inputs["y_substations"],
            allow_branching_approximation=self.options["approximate_branches"],
        ).to_csv(path_farm_location, index=False)

        self._orbit_config = config  # reinstall- probably not needed due to reference
        return config  # and return

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        """Creates and runs the project, then gathers the results."""

        # setup the custom-location library
        if self._path_library:
            initialize_library(self._path_library)

        # send it back to the superclass compute
        super().compute(
            inputs,
            outputs,
            discrete_inputs,
            discrete_outputs,
        )


class ORBITDetailedGroup(om.Group):
    """wrapper for ORBIT-WISDEM Fixed Substructure API, allowing manual IVC incorporation"""

    def initialize(self):
        """Initialize the group and declare options."""
        self.options.declare("case_title", default="working")
        self.options.declare(
            "modeling_options", types=dict, desc="Ard modeling options"
        )
        self.options.declare("approximate_branches", default=False)

    def setup(self):

        # add IVCs for landbosse
        variable_mapping = ORBIT_setup_latents(
            modeling_options=self.options["modeling_options"]
        )

        # create source independent variable components for LandBOSSE inputs
        for key, meta in variable_mapping.items():
            if key in ["number_of_turbines", "number_of_blades", "num_mooring_lines"]:
                comp = om.IndepVarComp()
                comp.add_discrete_output(name=key, val=meta["val"])
                self.add_subsystem(f"IVC_orbit_{key}", comp, promotes=["*"])
            else:
                self.add_subsystem(
                    f"IVC_orbit_{key}",
                    om.IndepVarComp(key, val=meta["val"], units=meta["units"]),
                    promotes=["*"],
                )

        # add orbit
        self.add_subsystem(
            "orbit",
            ORBITDetail(
                case_title=self.options["case_title"],
                modeling_options=self.options["modeling_options"],
                approximate_branches=self.options["approximate_branches"],
            ),
            promotes=[
                "total_capex",
                "total_capex_kW",
                "bos_capex",
                "installation_capex",
                "graph",
                "x_turbines",
                "y_turbines",
                "x_substations",
                "y_substations",
                # "plant_turbine_spacing",
                # "plant_row_spacing",
            ],
        )

        # connect
        for key in variable_mapping.keys():
            self.connect(key, f"orbit.{key}")



from contextlib import redirect_stdout, redirect_stderr
from functools import wraps
from io import StringIO
from pathlib import Path
import shutil
import sys

import openmdao.core.component


def extract_iter(component):
    """
    Extract the iter_count iff it exists, otherwise return None

    Extract the iteration count from a component's associated model.
    This function attempts to retrieve the iteration count from a component by
    traversing through its problem metadata and model reference. It safely
    handles cases where any of the required attributes or keys don't exist.

    Parameters
    ----------
        component: An object that may contain a _problem_meta attribute with
                  model reference information.
    Returns
    -------
        int or None: The iteration count from the model if it exists and is
                    accessible, otherwise None.
            The function returns None in the following cases:
            - component doesn't have a _problem_meta attribute
            - problem_meta doesn't contain a "model_ref" key
            - the model doesn't have an iter_count attribute
    """

    # extract the iter count if it exists, returning and handling none otherwise
    problem_meta = getattr(component, "_problem_meta", None)
    model = problem_meta.get("model_ref", lambda _: None)()
    iter_count = getattr(model, "iter_count", None)

    return iter_count


def get_storage_directory(
    component,
    storage_type: str = "logs",
    get_iter: bool = False,
    clean: bool = False,
):
    """
    Get a storage directory for the component constructed here.

    Take a component and create a storage directory (for, e.g. logs or init
    files), mirroring the OpenMDAO model structure as subdirectories, returning
    a pathlib.Path to the storage directory.

    Parameters
    ----------
    component : openmdao.core.Component
        an OpenMDAO component for which we want to create a storage directory
    storage_type : str, optional
        the type of storage sub-directory to make, by default "logs"
    get_iter : bool, optional
        should the storage directory tree be given an iteration subdirectory, by
        default False
    clean : bool, optional
        should the directory tree, if it already exists, be cleaned out, by
        default False

    Returns
    -------
    pathlib.Path
        the path to the storage subdirectory created
    """
    # the storage type we're doing (logs, discipline scripts, etc.)
    storage_dir = [
        storage_type,
    ]
    # if there's an iteration number to grab, grab it and add it to the dir
    iter = extract_iter(component) if get_iter else None
    if iter:
        storage_dir += [f"iter_{iter:04d}"]
    # mirror the comp path for a log directory
    subdir_logger = component.pathname.split(".")
    # find the reports directory
    dir_reports = Path(component._problem_meta["reports_dir"])
    # put the storage directory next to it
    path_storage = Path(dir_reports.parent, *storage_dir, *subdir_logger)

    # make a clean log location for this component if permitted
    try:
        path_storage.mkdir(parents=True, exist_ok=False)
    except FileExistsError:  # handle a FileExists, but raise anything else
        if clean:
            shutil.rmtree(path_storage, ignore_errors=True)
            path_storage.mkdir(parents=True, exist_ok=True)
        else:
            raise

    return path_storage


def name_create_log(component, iter: int = None):
    """
    For a given component, clean and create component- and rank-unique logfiles.

    Take a component and create logs, parallel to the reports file, mirroring
    the OpenMDAO model structure with stdout and stderr files for each rank,
    and finally return the file paths for the component to redirect stdout and
    stderr to.

    Parameters
    ----------
    component : openmdao.core.component.Component
        An OpenMDAO component that we want to capture stdout/stderr for

    Returns
    -------
    pathlib.Path
        a path to in the log system to dump stdout to
    pathlib.Path
        a path to in the log system to dump err to
    """

    # make sure we are dealing with an OM component
    if not isinstance(component, openmdao.core.component.Component):
        raise TypeError(
            f"Expected openmdao.core.component.Component, got {type(component)}"
        )

    path_logfile_template = (
        get_storage_directory(component, "logs", True, clean=True)
        / f"%s_rank{component._comm.rank:03d}.txt"
    )
    path_logfile_stdout = Path(path_logfile_template.as_posix() % "stdout")
    path_logfile_stderr = Path(path_logfile_template.as_posix() % "stderr")

    # return stdout and stderr files
    return path_logfile_stdout.absolute(), path_logfile_stderr.absolute()


def component_log_capture(compute_func, iter: int = None):
    """
    Decorator that redirects stdout and stderr to component-wise and rank-wise logfiles.

    This decorator will redirect stdout and stderr to component-wise and
    rank-wise logfiles, which are determined by the `name_create_log` function.
    The decorator uses context managers to redirect output streams to these
    files, ensuring that all print statements and errors within the function are
    logged appropriately.

    func : Callable
        The function to be decorated. It should be a method of a class, as
        `self` is expected as the first argument.

    Callable
        The wrapped function with stdout and stderr redirected to log files
        during its execution.
    """

    @wraps(compute_func)
    def wrapper(self, *args, **kwargs):

        # extract from modeling options the stdio_capture option iff it exists
        stdio_capture = getattr(self, "modeling_options", {}).get("stdio_capture")
        # bail out, returning the function w/ no changes if it doesn't
        if not stdio_capture:
            return compute_func(self, *args, **kwargs)

        # if we get here, we want to capture stdio

        # get log file paths
        path_stdout_log, path_stderr_log = name_create_log(self)

        try:
            # use context manager to redirect stdout & stderr
            with (
                open(path_stdout_log, "a") as stdout_file,
                open(path_stderr_log, "a") as stderr_file,
                redirect_stdout(stdout_file),
                redirect_stderr(stderr_file),
            ):
                return compute_func(self, *args, **kwargs)
        except Exception:
            raise  # make sure the exception is raised

    return wrapper


def prepend_tabs_to_stdio(func, tabs=1):
    @wraps(func)
    def wrapper(*args, **kwargs):
        old_stdout = sys.stdout
        sys.stdout = StringIO()

        # run the function
        returns = func(*args, **kwargs)

        # get capture output and restore stdout
        output = sys.stdout.getvalue()
        sys.stdout = old_stdout

        prefix = "\t"
        tabset = "".join([prefix for t in range(tabs)])
        if output:
            for line in output.splitlines():
                print(f"{tabset}{line}")

        # pass through the returns of the function
        return returns

    return wrapper



import numpy as np
import jax.numpy as jnp
import jax

jax.config.update("jax_enable_x64", True)
import ard.utils.geometry
import openmdao.api as om


class FarmExclusionDistancePolygon(om.ExplicitComponent):
    """
    A class to return distances between turbines and a polygonal exclusion, or
    sets of polygonal exclusion regions.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary

    Inputs
    ------
    x_turbines : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the turbines,
        with length `N_turbines` (mirrored w.r.t. `FarmAeroTemplate`)
    y_turbines : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the turbines,
        with length `N_turbines` (mirrored w.r.t. `FarmAeroTemplate`)

    Outputs
    -------
    exclusion_distances : np.ndarray
        a 1D array of distances (in meters) from each turbine to its assigned
        polygonal exclusion region
    """

    def initialize(self):
        """Initialization of the OpenMDAO component."""
        self.options.declare("modeling_options")

    def setup(self):
        """Setup of the OpenMDAO component."""

        # load modeling options
        self.modeling_options = self.options["modeling_options"]
        self.windIO = self.modeling_options["windIO_plant"]
        self.N_turbines = int(self.modeling_options["layout"]["N_turbines"])

        # load exclusion vertices from windIO file
        if "exclusions" not in self.windIO["site"]:
            raise KeyError(
                "You have requested an exclusion but no exclusions were found in the windIO file."
            )
        if "circle" in self.windIO["site"]["exclusions"]:
            raise NotImplementedError(
                "The circular exclusions from windIO have not been implemented here, yet."
            )
        if "polygons" not in self.windIO["site"]["exclusions"]:
            raise NotImplementedError(
                "Currently only polygon exclusions from windIO have been implemented and none were found."
            )
        self.exclusion_vertices = [
            np.array(
                [
                    polygon["x"],
                    polygon["y"],
                ]
            ).T
            for polygon in self.windIO["site"]["exclusions"]["polygons"]
        ]
        self.exclusion_regions = self.modeling_options.get("exclusions", {}).get(
            "turbine_exclusion_assignments",  # exclusion region assignments, if there
            np.zeros(self.N_turbines, dtype=int),  # default to zero for all turbines
        )

        # prep the jacobian
        self.distance_multi_point_to_multi_polygon_ray_casting_jac = jax.jacfwd(
            ard.utils.geometry.distance_multi_point_to_multi_polygon_ray_casting, [0, 1]
        )

        # set up inputs and outputs for turbine exclusion distances
        self.add_input(
            "x_turbines", jnp.zeros((self.N_turbines,)), units="m"
        )  # x location of the turbines in m w.r.t. reference coordinates
        self.add_input(
            "y_turbines", jnp.zeros((self.N_turbines,)), units="m"
        )  # y location of the turbines in m w.r.t. reference coordinates

        self.add_output(
            "exclusion_distances",
            jnp.zeros(self.N_turbines),
            units="m",
        )

    def setup_partials(self):
        """Derivative setup for the OpenMDAO component."""
        # override the OpenMDAO default FDM derivatives by declaring exact derivatives
        self.declare_partials(
            "*",
            "*",
            method="exact",
            rows=np.arange(self.N_turbines),
            cols=np.arange(self.N_turbines),
        )

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """Computation for the OpenMDAO component."""

        # unpack the working variables
        x_turbines = inputs["x_turbines"]
        y_turbines = inputs["y_turbines"]

        exclusion_distances = (
            ard.utils.geometry.distance_multi_point_to_multi_polygon_ray_casting(
                x_turbines,
                y_turbines,
                boundary_vertices=self.exclusion_vertices,
                regions=self.exclusion_regions,
            )
        )

        outputs["exclusion_distances"] = -exclusion_distances

    def compute_partials(self, inputs, partials, discrete_inputs=None):

        # unpack the working variables
        x_turbines = inputs["x_turbines"]
        y_turbines = inputs["y_turbines"]

        jacobian = self.distance_multi_point_to_multi_polygon_ray_casting_jac(
            x_turbines, y_turbines, self.exclusion_vertices, self.exclusion_regions
        )

        partials["exclusion_distances", "x_turbines"] = -jacobian[0].diagonal()
        partials["exclusion_distances", "y_turbines"] = -jacobian[1].diagonal()



import numpy as np

import openmdao.api as om


class LayoutTemplate(om.ExplicitComponent):
    """
    A template class for layout parametrizations.

    This is a template class that represents the fundamental input/output flow
    for a layout parametrization. This means outputting $(x,y)$ tuples (via
    `x_turbines` and `y_turbines) for the `N_turbines` in the farm, and
    outputting effective spacing metrics (for, e.g., simple BOS tools).

    Options
    -------
    modeling_options : dict
        a modeling options dictionary

    Inputs
    ------
    None

    Outputs
    -------
    x_turbines : np.ndarray
        a 1-D numpy array that represents that x (i.e. Easting) coordinate of
        the location of each of the turbines in the farm in meters
    y_turbines : np.ndarray
        a 1-D numpy array that represents that y (i.e. Northing) coordinate of
        the location of each of the turbines in the farm in meters
    spacing_effective_primary : float
        a measure of the spacing on a primary axis of a rectangular farm that
        would be equivalent to this one for the purposes of computing BOS costs
        measured in rotor diameters
    spacing_effective_secondary : float
        a measure of the spacing on a secondary axis of a rectangular farm that
        would be equivalent to this one for the purposes of computing BOS costs
        measured in rotor diameters
    """

    def initialize(self):
        """Initialization of OM component."""
        self.options.declare("modeling_options")

    def setup(self):
        """Setup of OM component."""

        # load modeling options
        modeling_options = self.modeling_options = self.options["modeling_options"]
        self.windIO = self.modeling_options["windIO_plant"]
        self.N_turbines = modeling_options["layout"]["N_turbines"]

        # add outputs that are universal
        self.add_output(
            "x_turbines",
            np.zeros((self.N_turbines,)),
            units="m",
            desc="turbine location in x-direction",
        )
        self.add_output(
            "y_turbines",
            np.zeros((self.N_turbines,)),
            units="m",
            desc="turbine location in y-direction",
        )
        self.add_output(
            "spacing_effective_primary",
            0.0,
            desc="effective spacing in x-dimension for BOS calculation",
        )
        self.add_output(
            "spacing_effective_secondary",
            0.0,
            desc="effective spacing in y-dimension for BOS calculation",
        )

    # omit setup partials for template class

    def compute(self, inputs, outputs):
        """
        Computation for the OM component.

        For a template class this is not implemented and raises an error!
        """

        raise NotImplementedError(
            "This is an abstract class for a derived class to implement"
        )


class LanduseTemplate(om.ExplicitComponent):
    """
    A template class for landuse computations.

    This is a template class that represents the fundamental input/output flow
    for a landuse calculation. Most details will be specialized based on use
    case, but most fundamentally it will intake a layback distance and output a
    simple area computation.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary

    Inputs
    ------
    distance_layback_diameters : float
        the number of diameters of layback desired for the landuse calculation

    Outputs
    -------
    area_tight : float
        the area in square kilometers that the farm occupies based on the
        circumscribing geometry with a specified (default zero) layback buffer
    """

    def initialize(self):
        """Initialization of OM component."""
        self.options.declare("modeling_options")

    def setup(self):
        """Setup of OM component."""

        # load modeling options and turbine count
        modeling_options = self.modeling_options = self.options["modeling_options"]
        self.windIO = self.modeling_options["windIO_plant"]
        self.N_turbines = modeling_options["layout"]["N_turbines"]

        # add inputs that are universal
        self.add_input(
            "distance_layback_diameters",
            0.0,
            units=None,
            desc="number of diameters of layback necessary for landuse",
        )

        # add outputs that are universal
        self.add_output(
            "area_tight",
            0.0,
            units="km**2",
            desc="fundamental area of the farm geometry",
        )

    # omit setup partials for template class

    def compute(self, inputs, outputs):
        """
        Computation for the OM component.

        For a template class this is not implemented and raises an error!
        """

        raise NotImplementedError(
            "This is an abstract class for a derived class to implement"
        )



import numpy as np
import jax.numpy as jnp
import jax


def smooth_max(x: jnp.ndarray, s: float = 1000.0) -> float:
    """Non-overflowing version of Smooth Max function (see ref 3 and 4 below).
    Calculates the smoothmax (a.k.a. softmax or LogSumExponential) of the elements in x.

    Based on implementation in BYU FLOW Lab's FLOWFarm software at
    (1) https://github.com/byuflowlab/FLOWFarm.jl/tree/master
    which is based on John D. Cook's writings at
    (2) https://www.johndcook.com/blog/2010/01/13/soft-maximum/
    and
    (3) https://www.johndcook.com/blog/2010/01/20/how-to-compute-the-soft-maximum/
    And based on article in FeedlyBlog
    (4) https://blog.feedly.com/tricks-of-the-trade-logsumexp/

    Args:
        x (list): list of values to be compared
        s (float, optional): alpha for smooth max function. Defaults to 1000.0.
            Larger values of `s` lead to more accurate results, but reduce the smoothness
            of the function.

    Returns:
        float: the smooth max of the provided `x` list
    """

    # get the maximum value and the index of maximum value
    max_ind = jnp.argmax(x)
    max_val = x[max_ind]

    # LogSumExp with smoothing factor s
    exponential = jnp.exp(
        s * (jnp.delete(x, max_ind, assume_unique_indices=True) - max_val)
    )
    r = (jnp.log(1.0 + jnp.sum(exponential)) + s * max_val) / s

    return r


smooth_max = jax.jit(smooth_max)


def smooth_min(x: np.ndarray, s: float = 1000.0) -> float:
    """Finds smooth min using the `smooth_max` function

    Args:
        x (list): list of values to be compared
        s (float, optional): alpha for smooth min function. Defaults to 1000.0.
            Larger values of `s` lead to more accurate results, but reduce the smoothness
            of the function.

    Returns:
        float: the smooth min of the provided `x` list
    """

    return -smooth_max(x=-x, s=s)


smooth_min = jax.jit(smooth_min)


def smooth_norm(vec: np.ndarray, buf: float = 1e-12) -> float:
    """Smooth version of the Frobenius, or 2, norm. This version is nearly equivalent to the 2-norm with the
    maximum absolute error corresponding to the order of the buffer value. The maximum error in the gradient is near unity, but
    the error in the gradient is generally about twice the error in the absolute value. The key benefit of the smooth norm is
    that it is differentiable at 0.0, while the standard norm is undefined at 0.0.

    Args:
        vec (np.ndarray): input vector to be normed
        buf (float, optional): buffer value included in the sum of squares part of the norm. Defaults to 1E-12.

    Returns:
        (float): normed result
    """
    return jnp.sqrt(buf**2 + jnp.sum(vec**2))


smooth_norm = jax.jit(smooth_norm)
smooth_norm_vec = jax.jit(jax.vmap(smooth_norm))



import numpy as np
from scipy.interpolate import RectBivariateSpline

import openmdao.api as om


class EagleDensityFunction(om.ExplicitComponent):
    """
    OpenMDAO component to evaluate eagle presence density at turbine locations.

    An Ard/OpenMDAO component that evaluates the eagle presence density metric
    calculated by the National Laboratory of the Rockies's Stochastic Soaring
    Raptor Simulator (SSRS) at the turbine locations. The eagle presence density
    is an output of an SSRS simulation indicating the unit density function of
    a raptor flying through the point during a given migratory period.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary (inherited from
        `templates.LanduseTemplate`)

    Inputs
    ------
    x_turbines : np.ndarray
        a 1-D numpy array that represents the x (i.e. Easting) coordinate of
        the location of each of the turbines in the farm in meters
    y_turbines : np.ndarray
        a 1-D numpy array that represents the y (i.e. Northing) coordinate of
        the location of each of the turbines in the farm in meters

    Outputs
    -------
    eagle_normalized_density : np.ndarray
        a 1-D numpy array that represents the normalized eagle presence density
        at each of the turbine locations (unitless)
    """

    def initialize(self):
        """Initialization of OM component."""
        self.options.declare("modeling_options")

    def setup(self):
        """Setup of OM component."""

        # load modeling options and turbine count
        modeling_options = self.modeling_options = self.options["modeling_options"]

        self.N_turbines = modeling_options["layout"]["N_turbines"]

        # grab the eagle presence density settings
        self.pres = self.modeling_options["eco"]["eagle_presence_density_map"]
        self.eagle_density_function = RectBivariateSpline(
            self.pres["x"], self.pres["y"], self.pres["normalized_presence_density"]
        )
        self.eagle_density_function_dx = self.eagle_density_function.partial_derivative(
            dx=1, dy=0
        )
        self.eagle_density_function_dy = self.eagle_density_function.partial_derivative(
            dx=0, dy=1
        )

        # add the full layout inputs
        self.add_input(
            "x_turbines",
            np.zeros((self.N_turbines,)),
            units="m",
            desc="turbine location in x-direction",
        )
        self.add_input(
            "y_turbines",
            np.zeros((self.N_turbines,)),
            units="m",
            desc="turbine location in y-direction",
        )

        # add outputs that are universal
        self.add_output(
            "eagle_normalized_density",
            np.zeros((self.N_turbines,)),
            units=None,
            desc="normalized eagle presence density",
        )

    def setup_partials(self):
        """Setup the OpenMDAO component partial derivatives."""
        self.declare_partials(
            "eagle_normalized_density",
            "x_turbines",
            diagonal=True,
            method="exact",
        )
        self.declare_partials(
            "eagle_normalized_density",
            "y_turbines",
            diagonal=True,
            method="exact",
        )

    def compute(self, inputs, outputs):
        """
        Computation for the OM component.
        """

        # unpack the turbine locations
        x_turbines = inputs["x_turbines"]  # m
        y_turbines = inputs["y_turbines"]  # m

        # evaluate the density function at each turbine point
        outputs["eagle_normalized_density"] = self.eagle_density_function(
            x_turbines,
            y_turbines,
            grid=False,
        )

    def compute_partials(self, inputs, partials):
        """
        Compute the partials for the OM component
        """

        # unpack the turbine locations
        x_turbines = inputs["x_turbines"]  # m
        y_turbines = inputs["y_turbines"]  # m

        # evaluate the gradients for each variable
        dfdx = self.eagle_density_function_dx(x_turbines, y_turbines, grid=False)
        dfdy = self.eagle_density_function_dy(x_turbines, y_turbines, grid=False)
        partials["eagle_normalized_density", "x_turbines"] = dfdx
        partials["eagle_normalized_density", "y_turbines"] = dfdy



import numpy as np
import matplotlib.pyplot as plt

import openmdao.api as om


class OutputLayout(om.ExplicitComponent):

    def initialize(self):
        self.options.declare("modeling_options")

    def setup(self):
        # load modeling options
        modeling_options = self.modeling_options = self.options["modeling_options"]
        self.N_turbines = modeling_options["farm"]["N_turbines"]

        # add inputs
        self.add_input(
            "x_turbines",
            np.zeros((self.N_turbines,)),
            units="m",
            desc="turbine location in x-direction",
        )
        self.add_input(
            "y_turbines",
            np.zeros((self.N_turbines,)),
            units="m",
            desc="turbine location in y-direction",
        )

    def compute(self, inputs, outputs):
        fig, ax = plt.subplots()

        ax.scatter(inputs["x_turbines"], inputs["y_turbines"])
        for idx, (x, y) in enumerate(zip(inputs["x_turbines"], inputs["y_turbines"])):
            ax.text(x, y, str(idx), ha="right", va="bottom")
        ax.axis("square")
        # plt.show()



import numpy as np


def get_plot_range(values, pct_buffer=5.0):
    """
    get the min and max values for a plot axis with a buffer applied

    Parameters
    ----------
    values : np.array
        the array of values in a given dimension
    pct_buffer : float, optional
        percent that should be included as a buffer, by default 5.0

    Returns
    -------
    float
        minimum value for the plot range
    float
        maximum value for the plot range
    """
    min_value = np.min(values)
    max_value = np.max(values)
    dvalues = max_value - min_value
    min_value = min_value - pct_buffer / 100.0 * dvalues
    max_value = max_value + pct_buffer / 100.0 * dvalues
    return min_value, max_value



from pathlib import Path

import numpy as np

import openmdao.api as om

import floris


def create_windresource_from_windIO(
    windIOdict: dict,
    resource_type: str = None,  # ["probability", "timeseries", "weibull_sector"]
):
    """
    takes a windIO plant specification and creates an appropriate wind resource

    Parameters
    ----------
    windIOdict : dict
        a full, presumed validated, windIO plant specification file
    resource_type : str, optional
        one of "probability", "timeseries", "weibull_sector" indicating, either
        a "probability"-based representation by a FLORIS WindRose object, a
        "timeseries" representation using a FLORIS TimeSeries object, or a
        "weibull_sector" representation which has not yet been implemented

    Returns
    -------
    floris.WindDataBase
        a FLORIS wind resource object that encodes the windIO wind resource

    Raises
    ------
    KeyError
        if the input file is not read correctly
    ValueError
        if values found in the input file have an issue
    NotImplementedError
        if an unimplemented case is found
    """

    if not "site" in windIOdict:  # make sure the site is specified
        raise KeyError("No site specified in windIO plant dictionary.")
    if "energy_resource" not in windIOdict["site"]:
        raise KeyError("Missing 'energy_resource' in windIOdict['site'].")
    if "wind_resource" not in windIOdict["site"]["energy_resource"]:
        raise KeyError(
            "Missing 'wind_resource' in windIOdict['site']['energy_resource']."
        )

    # get the wind resource specification out of the dictionary
    wind_resource = windIOdict["site"]["energy_resource"]["wind_resource"]

    # figure out the case in play
    fields_wind_resource = wind_resource.keys()
    case_probability_based = all(
        val in fields_wind_resource
        for val in ["probability", "wind_direction", "wind_speed"]
    )
    case_weibull_based = all(
        val in fields_wind_resource
        for val in ["weibull_a", "weibull_k", "weibull_probability"]
    )
    case_timeseries_based = all(
        val in fields_wind_resource for val in ["time", "wind_direction", "wind_speed"]
    )

    if case_weibull_based and not (case_probability_based or case_timeseries_based):
        if resource_type is not None and resource_type != "weibull_sector":
            raise ValueError(
                f"Attempted to load {resource_type}-type wind resource and "
                "only weibull_sector was found."
            )
        raise NotImplementedError(
            "Sector Weibull wind resource has not yet been implemented for FLORIS."
        )

    wind_resource_representation = None

    if case_probability_based and case_timeseries_based:
        raise ValueError(
            "Both probability-based and time-series-based wind resource "
            "specifications have been found; this is ambiguous."
        )
    elif case_probability_based:
        if resource_type is not None and resource_type != "probability":
            raise ValueError(
                f"Attempted to load {resource_type}-type wind resource and "
                "only probability was found."
            )

        # extract key variables
        wind_directions = np.array(
            wind_resource["wind_direction"]["data"]
            if "data" in wind_resource["wind_direction"]
            else wind_resource["wind_direction"]
        )
        wind_speeds = np.array(
            wind_resource["wind_speed"]["data"]
            if "data" in wind_resource["wind_speed"]
            else wind_resource["wind_speed"]
        )
        probabilities = np.array(wind_resource["probability"]["data"])

        if "turbulence_intensity" not in wind_resource:
            raise KeyError(
                "windIO does not require turbulence intensities to be set, but "
                "FLORIS requires turbulence intensities; please set the "
                "turbulence intensities in the windIO file."
            )
        turbulence_intensities = np.array(
            wind_resource["turbulence_intensity"]["data"]
            if "data" in wind_resource["turbulence_intensity"]
            else wind_resource["turbulence_intensity"]
        )

        # create FLORIS representation
        wind_resource_representation = floris.WindRose(
            wind_directions=wind_directions,
            wind_speeds=wind_speeds,
            freq_table=probabilities,
            ti_table=turbulence_intensities,
        )
        # stash some metadata for the wind resource
        wind_resource_representation.reference_height = (
            wind_resource["reference_height"]
            if "reference_height" in wind_resource
            else None
        )

        return wind_resource_representation

    elif case_timeseries_based:
        if resource_type is not None and resource_type != "timeseries":
            raise ValueError(
                f"Attempted to load {resource_type}-type wind resource and "
                "only time-series was found."
            )

        wind_directions = np.array(
            wind_resource["wind_direction"].get("data", wind_resource["wind_direction"])
            if type(wind_resource["wind_direction"]) is dict
            else wind_resource["wind_direction"]
        )
        wind_speeds = np.array(
            wind_resource["wind_speed"].get("data", wind_resource["wind_speed"])
            if type(wind_resource["wind_speed"]) is dict
            else wind_resource["wind_speed"]
        )
        if "turbulence_intensity" in wind_resource:
            if "data" in wind_resource["turbulence_intensity"]:
                turbulence_intensities = np.array(
                    wind_resource["turbulence_intensity"]["data"]
                )
            else:
                turbulence_intensities = np.array(wind_resource["turbulence_intensity"])
        else:
            raise KeyError(
                "Missing 'turbulence_intensity' in time-series wind resource."
            )

        wind_resource_representation = floris.TimeSeries(
            wind_directions=wind_directions,
            wind_speeds=wind_speeds,
            turbulence_intensities=turbulence_intensities,
        )
        # stash some metadata for the wind resource
        wind_resource_representation.reference_height = (
            wind_resource["reference_height"]
            if "reference_height" in wind_resource
            else None
        )
        wind_resource_representation.time = (
            wind_resource["time"] if "time" in wind_resource else None
        )

        return wind_resource_representation

    else:
        raise ValueError(
            "You shouldn't have ended here, try validating the windIO yaml file."
        )


class FarmAeroTemplate(om.ExplicitComponent):
    """
    Template component for using a farm aerodynamics model.

    A farm aerodynamics component, based on this template, will compute the
    aerodynamics for a farm with some layout and yaw configuration.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary

    Inputs
    ------
    x_turbines : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the turbines,
        with length `N_turbines`
    y_turbines : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the turbines,
        with length `N_turbines`
    yaw_turbines : np.ndarray
        a numpy array indicating the yaw angle to drive each turbine to with
        respect to the ambient wind direction, with length `N_turbines`

    Outputs
    -------
    None
    """

    def initialize(self):
        """Initialization of OM component."""
        self.options.declare("modeling_options")
        self.options.declare("data_path")

    def setup(self):
        """Setup of OM component."""

        # load modeling options
        self.modeling_options = self.options["modeling_options"]
        self.windIO = self.modeling_options["windIO_plant"]
        self.N_turbines = self.modeling_options["layout"]["N_turbines"]

        # set up inputs and outputs for farm layout
        self.add_input("x_turbines", np.zeros((self.N_turbines,)), units="m")
        self.add_input("y_turbines", np.zeros((self.N_turbines,)), units="m")
        self.add_input(
            "yaw_turbines",
            np.zeros((self.N_turbines,)),
            units="deg",
        )

    def compute(self, inputs, outputs):
        """
        Computation for the OM component.

        For a template class this is not implemented and raises an error!
        """

        #############################################
        #                                           #
        # IMPLEMENT THE AERODYNAMICS COMPONENT HERE #
        #                                           #
        #############################################

        raise NotImplementedError(
            "This is an abstract class for a derived class to implement!"
        )


class BatchFarmPowerTemplate(FarmAeroTemplate):
    """
    Template component for computing power using a farm aerodynamics model.

    A farm power component, based on this template, will compute the power and
    thrust for a farm composed of a given rotor type.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary (inherited from `FarmAeroTemplate`)

    Inputs
    ------
    x_turbines : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the turbines,
        with length `N_turbines` (inherited from `FarmAeroTemplate`)
    y_turbines : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the turbines,
        with length `N_turbines` (inherited from `FarmAeroTemplate`)
    yaw_turbines : np.ndarray
        a numpy array indicating the yaw angle to drive each turbine to with
        respect to the ambient wind direction, with length `N_turbines`
        (inherited from `FarmAeroTemplate`)

    Outputs
    -------
    power_farm : np.ndarray
        an array of the farm power for each of the wind conditions that have
        been queried
    power_turbines : np.ndarray
        an array of the farm power for each of the turbines in the farm across
        all of the conditions that have been queried on the wind rose
        (`N_turbines`, `N_wind_conditions`)
    thrust_turbines : np.ndarray
        an array of the wind turbine thrust for each of the turbines in the farm
        across all of the conditions that have been queried on the wind rose
        (`N_turbines`, `N_wind_conditions`)
    """

    def initialize(self):
        """Initialization of OM component."""
        super().initialize()

    def setup(self):
        """Setup of OM component."""
        super().setup()

        # unpack wind query object
        self.wind_query = create_windresource_from_windIO(
            self.windIO,
            "timeseries",
        )
        self.directions_wind = self.wind_query.wind_directions.tolist()
        self.speeds_wind = self.wind_query.wind_speeds.tolist()
        self.TIs_wind = self.wind_query.turbulence_intensities.tolist()
        self.N_wind_conditions = len(self.directions_wind)

        # add the outputs we want for a batched power analysis:
        #   - farm and turbine powers
        #   - turbine thrusts

        # self.add_output(
        #     "AEP_farm",
        #     np.array([0.0]),
        #     units="W*h",
        # )
        self.add_output(
            "AEP_farm",
            0.0,
            units="W*h",
        )
        self.add_output(
            "power_farm",
            np.zeros((self.N_wind_conditions,)),
            units="W",
        )

        if (
            self.options["modeling_options"]
            .get("aero", {})
            .get("return_turbine_output")
        ):
            self.add_output(
                "power_turbines",
                np.zeros((self.N_turbines, self.N_wind_conditions)),
                units="W",
            )
            self.add_output(
                "thrust_turbines",
                np.zeros((self.N_turbines, self.N_wind_conditions)),
                units="N",
            )

    def setup_partials(self):
        """Derivative setup for OM component."""
        # the default (but not preferred!) derivatives are FDM
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        """
        Computation for the OM component.

        For a template class this is not implemented and raises an error!
        """

        #############################################
        #                                           #
        # IMPLEMENT THE AERODYNAMICS COMPONENT HERE #
        #                                           #
        #############################################

        raise NotImplementedError(
            "This is an abstract class for a derived class to implement!"
        )

        # the following should be set
        outputs["power_farm"] = np.zeros((self.N_wind_conditions,))
        if (
            self.options["modeling_options"]
            .get("aero", {})
            .get("return_turbine_output")
        ):
            outputs["power_turbines"] = np.zeros(
                (self.N_turbines, self.N_wind_conditions)
            )
            outputs["thrust_turbines"] = np.zeros(
                (self.N_turbines, self.N_wind_conditions)
            )


class FarmAEPTemplate(FarmAeroTemplate):
    """
    A template component for computing power using a farm aerodynamics model.

    A farm power component, based on this template, will compute the power and
    thrust for a farm composed of a given rotor type.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary (inherited from FarmAeroTemplate)
    data_path: str
        absolute path to data directory

    Inputs
    ------
    x_turbines : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the turbines,
        with length `N_turbines` (inherited from `FarmAeroTemplate`)
    y_turbines : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the turbines,
        with length `N_turbines` (inherited from `FarmAeroTemplate`)
    yaw_turbines : np.ndarray
        a numpy array indicating the yaw angle to drive each turbine to with
        respect to the ambient wind direction, with length `N_turbines`
        (inherited from `FarmAeroTemplate`)

    Outputs
    -------
    AEP_farm : float
        the AEP of the farm given by the analysis
    power_farm : np.ndarray
        an array of the farm power for each of the wind conditions that have
        been queried
    power_turbines : np.ndarray
        an array of the farm power for each of the turbines in the farm across
        all of the conditions that have been queried on the wind rose
        (`N_turbines`, `N_wind_conditions`)
    thrust_turbines : np.ndarray
        an array of the wind turbine thrust for each of the turbines in the farm
        across all of the conditions that have been queried on the wind rose
        (`N_turbines`, `N_wind_conditions`)
    """

    def initialize(self):
        """Initialization of OM component."""
        super().initialize()
        self.options.declare("data_path", default=None)

    def setup(self):
        """Setup of OM component."""
        super().setup()
        data_path = str(self.options["data_path"])

        self.wind_query = create_windresource_from_windIO(
            self.windIO,
            "probability",
        )

        if data_path is None:
            data_path = ""

        self.directions_wind = self.wind_query.wind_directions
        self.speeds_wind = self.wind_query.wind_speeds
        self.TIs_wind = self.wind_query.ti_table_flat
        self.pmf_wind = self.wind_query.freq_table_flat
        self.N_wind_conditions = len(self.pmf_wind)

        # add the outputs we want for an AEP analysis:
        #   - AEP estimate
        #   - farm and turbine powers
        #   - turbine thrusts
        self.add_output(
            "AEP_farm",
            0.0,
            units="W*h",
        )
        self.add_output(
            "power_farm",
            np.zeros((self.N_wind_conditions,)),
            units="W",
        )
        if (
            self.options["modeling_options"]
            .get("aero", {})
            .get("return_turbine_output")
        ):
            self.add_output(
                "power_turbines",
                np.zeros((self.N_turbines, self.N_wind_conditions)),
                units="W",
            )
            self.add_output(
                "thrust_turbines",
                np.zeros((self.N_turbines, self.N_wind_conditions)),
                units="N",
            )
        # ... more outputs can be added here

    def setup_partials(self):
        """Derivative setup for OM component."""
        # the default (but not preferred!) derivatives are FDM
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        """
        Computation for the OM component.

        For a template class this is not implemented and raises an error!
        """

        #############################################
        #                                           #
        # IMPLEMENT THE AERODYNAMICS COMPONENT HERE #
        #                                           #
        #############################################

        raise NotImplementedError(
            "This is an abstract class for a derived class to implement!"
        )

        # the following should be set
        outputs["AEP_farm"] = 0.0
        outputs["power_farm"] = np.zeros((self.N_wind_conditions,))
        outputs["power_turbines"] = np.zeros((self.N_turbines, self.N_wind_conditions))
        outputs["thrust_turbines"] = np.zeros((self.N_turbines, self.N_wind_conditions))



import numpy as np

import openmdao.api as om

import math

import ard.geographic


def generate_anchor_points(
    center: np.ndarray, length: float, rotation_deg: float, N: int
) -> np.ndarray:
    """Generates anchor points equally spaced around the platform

    Args:
        center (np.ndarray): x and y of platform in km
        length (float): desired horizontal anchor length in km
        rotation_deg (float): rotation in deg. counter-clockwise from east
        N (int): number of anchors

    Returns:
        np.ndarray: array of size N by 2 containing the x and y positions of each anchor
    """

    cx, cy = center
    angle_step = 360 / N
    lines = np.zeros([N, 2])

    for i in range(N):
        angle_deg = rotation_deg + i * angle_step
        angle_rad = math.radians(angle_deg)
        x = cx + length * math.cos(angle_rad)
        y = cy + length * math.sin(angle_rad)
        lines[i, 0] = x
        lines[i, 1] = y

    return lines


def simple_mooring_design(
    phi_platform: np.ndarray,
    x_turbines: np.ndarray,
    y_turbines: np.ndarray,
    length: float,
    N_turbines: int,
    N_anchors: int,
) -> tuple[np.ndarray]:
    """_summary_

    Args:
        phi_platform (np.ndarray): counterclockwise rotation from east in deg. for each platform
        x_turbines (np.ndarray): list of platform/turbine easting in km
        y_turbines (np.ndarray): list of platform/turbine northing in km
        length (float): desired horizontal anchor length in km
        N_turbines (int): number of wind turbines in the farm
        N_anchors (int): number of anchors per turbine/platform

    Returns:
        tuple[np.ndarray]: x locations of anchors, y locations of anchors, each array of shape N_turbines by N_anchors
    """

    x_anchors = np.zeros([N_turbines, N_anchors])
    y_anchors = np.zeros_like(x_anchors)

    for i, (x, y) in enumerate(zip(x_turbines, y_turbines)):

        center = (x, y)

        anchors = generate_anchor_points(
            center=center, length=length, rotation_deg=phi_platform[i], N=N_anchors
        )

        for j in range(N_anchors):
            x_anchors[i, j] = anchors[j, 0]
            y_anchors[i, j] = anchors[j, 1]

    return x_anchors, y_anchors


class ConstantDepthMooringDesign(om.ExplicitComponent):
    """
    A class to create a constant-depth simplified mooring design for a floating
    offshore wind farm.

    This is a class that should be used to generate a floating offshore wind
    farm's collective mooring system.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary (inherited from `FarmAeroTemplate`)
    wind_query : floris.wind_data.WindRose
        a WindQuery objects that specifies the wind conditions that are to be
        computed
    bathymetry_data : ard.geographic.BathymetryData
        a BathymetryData object to specify the bathymetry mesh/sampling

    Inputs
    ------
    phi_platform : np.ndarray
        a 1D numpy array indicating the cardinal direction angle of the mooring
        orientation, with length `N_turbines`
    x_turbines : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the turbines,
        with length `N_turbines` (mirrored w.r.t. `FarmAeroTemplate`)
    y_turbines : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the turbines,
        with length `N_turbines` (mirrored w.r.t. `FarmAeroTemplate`)
    thrust_turbines : np.ndarray
        an array of the wind turbine thrust for each of the turbines in the farm
        across all of the conditions that have been queried on the wind rose
        (`N_turbines`, `N_wind_conditions`)

    Outputs
    -------
    x_anchors : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the mooring
        system anchors, with shape `N_turbines` x `N_anchors`
    y_anchors : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the mooring
        system anchors, with shape `N_turbines` x `N_anchors`

    """

    def initialize(self):
        """Initialization of the OpenMDAO component."""
        self.options.declare("modeling_options")

        # farm power wind conditions query (not necessarily a full wind rose)
        self.options.declare("wind_query")

        # currently I'm thinking of sea bed conditions as a class, see above
        self.options.declare("bathymetry_data")  # BatyhmetryData object

    def setup(self):
        """Setup of the OpenMDAO component."""

        # load modeling options
        self.modeling_options = self.options["modeling_options"]
        self.N_turbines = self.modeling_options["layout"]["N_turbines"]
        self.N_anchors = self.modeling_options["platform"]["N_anchors"]
        self.min_mooring_line_length_m = self.modeling_options["platform"][
            "min_mooring_line_length_m"
        ]

        # get the number of wind conditions (for thrust measurements)
        if self.options["wind_query"] is not None:
            self.N_wind_conditions = self.options["wind_query"].N_conditions
        # MANAGE ADDITIONAL LATENT VARIABLES HERE!!!!!

        # set up inputs and outputs for mooring system
        self.add_input(
            "phi_platform", np.zeros((self.N_turbines,)), units="deg"
        )  # cardinal direction of the mooring platform orientation
        self.add_input(
            "x_turbines", np.zeros((self.N_turbines,)), units="km"
        )  # x location of the mooring platform in km w.r.t. reference coordinates
        self.add_input(
            "y_turbines", np.zeros((self.N_turbines,)), units="km"
        )  # y location of the mooring platform in km w.r.t. reference coordinates
        if self.options["wind_query"] is not None:
            self.add_input(
                "thrust_turbines",
                np.zeros((self.N_turbines, self.N_wind_conditions)),
                units="kN",
            )  # turbine thrust coming from each wind direction
        # ADD ADDITIONAL (DESIGN VARIABLE) INPUTS HERE!!!!!

        self.add_output(
            "x_anchors",
            np.zeros((self.N_turbines, self.N_anchors)),
            units="km",
        )  # x location of the mooring platform in km w.r.t. reference coordinates
        self.add_output(
            "y_anchors",
            np.zeros((self.N_turbines, self.N_anchors)),
            units="km",
        )  # y location of the mooring platform in km w.r.t. reference coordinates

    def setup_partials(self):
        """Derivative setup for the OpenMDAO component."""
        # the default (but not preferred!) derivatives are FDM
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """Computation for the OpenMDAO component."""

        # unpack the working variables
        phi_platform = inputs["phi_platform"]
        x_turbines = inputs["x_turbines"]
        y_turbines = inputs["y_turbines"]
        # thrust_turbines = inputs["thrust_turbines"]  #

        # BEGIN: REPLACE ME

        x_anchors, y_anchors = simple_mooring_design(
            phi_platform=phi_platform,
            x_turbines=x_turbines,
            y_turbines=y_turbines,
            length=self.min_mooring_line_length_m * 1e-3,  # convert to km
            N_turbines=self.N_turbines,
            N_anchors=self.N_anchors,
        )

        # END REPLACE ME

        # replace the below with the final anchor locations...
        outputs["x_anchors"] = x_anchors
        outputs["y_anchors"] = y_anchors



import numpy as np

import ard.layout.templates as templates


class GridFarmLayout(templates.LayoutTemplate):
    """
    A simplified, uniform four-parameter parallelepiped grid farm layout class.

    This is a class to take a parameterized, structured grid farm defined by a
    gridded parallelepiped with spacing variables defined to:
    1) orient the farm with respect to North,
    2) space the rows of turbines along this primary vector,
    3) space the columns of turbines along the perpendicular, and
    4) skew the positioning along a parallel to the primary (orientation) vector.

    The layout model is shown in a ASCII image below:

    ::

    |                                          |-------| <- streamwise spacing
    |          orient.         x ----- x ----- x ----- x ----- x -
    |           angle         /       /       /       /       /  | <- spanwise spacing
    |            |           x ----- x ----- x ----- x ----- x   -      (perpendicular
    |            v          /       /       /       /       /            w.r.t. primary)
    |           -------    x ----- x ----- x ----- x ----- x    ----- primary vector
    |                '    /       /       /       /       /             (rotated from
    |            '       x ----- x ----- x ----- x ----- x               north CW by
    |        '          /       /       /       /       /                orientation
    |        NORTH     x ----- x ----- x ----- x ----- x                 angle)
    |                        /|
    |                       / |
    |                      /  | <- skew angle


    Options
    -------
    modeling_options : dict
        a modeling options dictionary (inherited from
        `templates.LayoutTemplate`)
    N_turbines : int
        the number of turbines that should be in the farm layout (inherited from
        `templates.LayoutTemplate`)

    Inputs
    ------
    angle_orientation : float
        orientation in degrees clockwise with respect to North of the primary
        axis of the wind farm layout
    spacing_primary : float
        spacing of turbine rows along the primary axis (rotated by
        `angle_orientation`) in nondimensional rotor diameters
    spacing_secondary : float
        spacing of turbine columns along the perpendicular to the primary axis
        (rotated by 90° with respect to the primary axis) in nondimensional
        rotor diameters
    angle_skew : float
        clockwise skew angle of turbine rows w.r.t. beyond the 90° clockwise
        perpendicular to the primary axis

    Outputs
    -------
    x_turbines : np.ndarray
        a 1-D numpy array that represents that x (i.e. Easting) coordinate of
        the location of each of the turbines in the farm in meters (inherited
        from `templates.LayoutTemplate`)
    y_turbines : np.ndarray
        a 1-D numpy array that represents that y (i.e. Northing) coordinate of
        the location of each of the turbines in the farm in meters (inherited
        from `templates.LayoutTemplate`)
    spacing_effective_primary : float
        a measure of the spacing on a primary axis of a rectangular farm that
        would be equivalent to this one for the purposes of computing BOS costs
        measured in rotor diameters (inherited from `templates.LayoutTemplate`)
    spacing_effective_secondary : float
        a measure of the spacing on a secondary axis of a rectangular farm that
        would be equivalent to this one for the purposes of computing BOS costs
        measured in rotor diameters (inherited from `templates.LayoutTemplate`)
    """

    def initialize(self):
        """Initialization of OM component."""
        super().initialize()

    def setup(self):
        """Setup of OM component."""
        super().setup()
        spacing_primary = self.modeling_options["layout"]["spacing_primary"]
        spacing_secondary = self.modeling_options["layout"]["spacing_secondary"]
        angle_orientation = self.modeling_options["layout"]["angle_orientation"]
        angle_skew = self.modeling_options["layout"]["angle_skew"]

        # add four-parameter grid farm layout DVs
        self.add_input("spacing_primary", spacing_primary, units=None)
        self.add_input("spacing_secondary", spacing_secondary, units=None)
        self.add_input("angle_orientation", angle_orientation, units="deg")
        self.add_input("angle_skew", angle_skew, units="deg")

    def setup_partials(self):
        """Derivative setup for OM component."""

        # default FD for the layout tools
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        """Computation for the OM component."""

        D_rotor = self.windIO["wind_farm"]["turbine"][
            "rotor_diameter"
        ]  # will break if multiple turbine types are used...
        lengthscale_spacing_streamwise = inputs["spacing_primary"] * D_rotor
        lengthscale_spacing_spanwise = inputs["spacing_secondary"] * D_rotor

        N_square = int(np.sqrt(self.N_turbines))  # floors

        count_y, count_x = np.meshgrid(
            np.arange(-((N_square - 1) / 2), ((N_square + 1) / 2)),
            np.arange(-((N_square - 1) / 2), ((N_square + 1) / 2)),
        )

        if self.N_turbines == N_square**2:
            pass
        elif self.N_turbines <= N_square * (N_square + 1):
            # grid farm is a little bit above the last square... add a trailing
            # row.
            count_x = np.vstack([count_x, ((N_square + 1) / 2) * np.ones((N_square,))])
            count_y = np.vstack(
                [count_y, np.arange(-((N_square - 1) / 2), ((N_square + 1) / 2))]
            )
            count_x = count_x.flatten()
            count_y = count_y.flatten()
        else:
            # grid farm is nearly the next square... oversize and cut the last
            count_y, count_x = np.meshgrid(
                np.arange(-((N_square) / 2), ((N_square + 2) / 2)),
                np.arange(-((N_square) / 2), ((N_square + 2) / 2)),
            )
        count_x = count_x.flatten()[: self.N_turbines]
        count_y = count_y.flatten()[: self.N_turbines]

        angle_skew = -np.pi / 180.0 * inputs["angle_skew"]
        Bmtx = np.array(
            [
                [1.0, 0.0],
                [np.tan(angle_skew[0]), 1.0],
            ]
        ).squeeze()

        xi_positions = count_x * lengthscale_spacing_spanwise
        yi_positions = count_y * lengthscale_spacing_streamwise

        angle_orientation = np.pi / 180.0 * inputs["angle_orientation"]
        Amtx = np.array(
            [
                [np.cos(angle_orientation), np.sin(angle_orientation)],
                [-np.sin(angle_orientation), np.cos(angle_orientation)],
            ]
        ).squeeze()
        xyp = Amtx @ (Bmtx @ np.vstack([xi_positions, yi_positions]))

        outputs["x_turbines"] = xyp[0, :].tolist()
        outputs["y_turbines"] = xyp[1, :].tolist()

        outputs["spacing_effective_primary"] = inputs["spacing_primary"]
        outputs["spacing_effective_secondary"] = np.sqrt(
            inputs["spacing_secondary"] ** 2.0 / np.cos(angle_skew) ** 2.0
        )


class GridFarmLanduse(templates.LanduseTemplate):
    """
    Landuse class for four-parameter parallelepiped grid farm layout.

    This is a class to compute the landuse area of the parameterized, structured
    grid farm defined in `GridFarmLayout`.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary (inherited from
        `templates.LayoutTemplate`)
    N_turbines : int
        the number of turbines that should be in the farm layout (inherited from
        `templates.LayoutTemplate`)

    Inputs
    ------
    distance_layback_diameters : float
        the number of diameters of layback desired for the landuse calculation
        (inherited from `templates.LayoutTemplate`)
    angle_orientation : float
        orientation in degrees clockwise with respect to North of the primary
        axis of the wind farm layout
    spacing_primary : float
        spacing of turbine rows along the primary axis (rotated by
        `angle_orientation`) in nondimensional rotor diameters
    spacing_secondary : float
        spacing of turbine columns along the perpendicular to the primary axis
        (rotated by 90° with respect to the primary axis) in nondimensional
        rotor diameters
    angle_skew : float
        clockwise skew angle of turbine rows w.r.t. beyond the 90° clockwise
        perpendicular to the primary axis

    Outputs
    -------
    area_tight : float
        the area in square kilometers that the farm occupies based on the
        circumscribing geometry with a specified (default zero) layback buffer
        (inherited from `templates.LayoutTemplate`)
    area_aligned_parcel : float
        the area in square kilometers that the farm occupies based on the
        circumscribing rectangle that is aligned with the primary axis of the
        wind farm plus a specified (default zero) layback buffer
    area_compass_parcel : float
        the area in square kilometers that the farm occupies based on the
        circumscribing rectangle that is aligned with the compass rose plus a
        specified (default zero) layback buffer
    """

    def setup(self):
        """Setup of OM component."""

        super().setup()

        # add grid farm-specific inputs
        self.add_input(
            "spacing_primary",
            self.modeling_options["layout"]["spacing_primary"],
            units=None,
            desc="turbine row spacing in rotor diameters",
        )
        self.add_input(
            "spacing_secondary",
            self.modeling_options["layout"]["spacing_secondary"],
            units=None,
            desc="turbine column spacing (along rows) in rotor diameters",
        )
        self.add_input(
            "angle_orientation",
            self.modeling_options["layout"]["angle_orientation"],
            units="deg",
            desc="orientation in degrees clockwise with respect to North",
        )
        self.add_input(
            "angle_skew",
            self.modeling_options["layout"]["angle_skew"],
            units="deg",
            desc="clockwise skew angle of turbine rows",
        )

        self.add_output(
            "area_aligned_parcel",
            0.0,
            units="km**2",
            desc="area of the tightest rectangle around the farm (plus layback) that is aligned with the orientation vector",
        )
        self.add_output(
            "area_compass_parcel",
            0.0,
            units="km**2",
            desc="area of the tightest rectangular and compass-aligned land parcel that will fit the farm (plus layback)",
        )

    def setup_partials(self):
        """Derivative setup for OM component."""

        # default complex step for the layout tools, since they're often algebraic
        self.declare_partials("*", "*", method="cs")

    def compute(self, inputs, outputs):
        """Computation for the OM component."""

        D_rotor = self.windIO["wind_farm"]["turbine"][
            "rotor_diameter"
        ]  # will break if multiple turbine types are used...
        lengthscale_spacing_streamwise = inputs["spacing_primary"] * D_rotor
        lengthscale_spacing_spanwise = inputs["spacing_secondary"] * D_rotor
        lengthscale_layback = inputs["distance_layback_diameters"] * D_rotor

        N_square = int(np.sqrt(self.N_turbines))  # floors

        min_count_xf = -(N_square - 1) / 2
        min_count_yf = -(N_square - 1) / 2
        max_count_xf = (N_square - 1) / 2
        max_count_yf = (N_square - 1) / 2

        if self.N_turbines == N_square**2:
            pass
        elif self.N_turbines <= N_square * (N_square + 1):
            max_count_xf = (N_square + 1) / 2
        else:
            min_count_xf = -(N_square) / 2
            min_count_yf = -(N_square) / 2
            max_count_xf = (N_square) / 2
            max_count_yf = (N_square) / 2

        # the side lengths of a parallelopiped oriented with the farm that encloses the farm with layback
        length_farm_xf = (max_count_xf - min_count_xf) * lengthscale_spacing_spanwise
        length_farm_yf = (max_count_yf - min_count_yf) * lengthscale_spacing_streamwise

        # the area of a parallelopiped oriented with the farm that encloses the farm with layback
        area_parallelopiped = (length_farm_xf + 2 * lengthscale_layback) * (
            length_farm_yf + 2 * lengthscale_layback
        )

        # the side lengths of a square oriented with the farm that encloses the farm with layback
        angle_skew = np.pi / 180.0 * inputs["angle_skew"]
        length_enclosing_farm_xf = length_farm_xf
        length_enclosing_farm_yf = (
            max_count_yf * lengthscale_spacing_streamwise
            + np.abs(max_count_xf)
            * lengthscale_spacing_spanwise
            * np.abs(np.tan(angle_skew))
        ) - (
            min_count_yf * lengthscale_spacing_streamwise
            - np.abs(min_count_xf)
            * lengthscale_spacing_spanwise
            * np.abs(np.tan(angle_skew))
        )

        # the area of a square oriented with the farm that encloses the farm with layback
        area_enclosingsquare_farmoriented = (
            length_enclosing_farm_xf + 2 * lengthscale_layback
        ) * (length_enclosing_farm_yf + 2 * lengthscale_layback)

        # the side lengths of a square oriented with the compass rose that encloses the farm with layback
        angle_orientation = np.pi / 180.0 * inputs["angle_orientation"]
        A_x = (
            +max_count_yf * lengthscale_spacing_streamwise * np.sin(angle_orientation)
            + max_count_xf * lengthscale_spacing_spanwise * np.cos(angle_orientation)
            - (max_count_xf * lengthscale_spacing_spanwise * np.sin(angle_orientation))
            * np.tan(angle_skew)
        )
        A_y = (
            +max_count_yf * lengthscale_spacing_streamwise * np.cos(angle_orientation)
            - max_count_xf * lengthscale_spacing_spanwise * np.sin(angle_orientation)
            - (max_count_xf * lengthscale_spacing_spanwise * np.cos(angle_orientation))
            * np.tan(angle_skew)
        )
        B_x = (
            +min_count_yf * lengthscale_spacing_streamwise * np.sin(angle_orientation)
            + max_count_xf * lengthscale_spacing_spanwise * np.cos(angle_orientation)
            - (max_count_xf * lengthscale_spacing_spanwise * np.sin(angle_orientation))
            * np.tan(angle_skew)
        )
        B_y = (
            +min_count_yf * lengthscale_spacing_streamwise * np.cos(angle_orientation)
            - max_count_xf * lengthscale_spacing_spanwise * np.sin(angle_orientation)
            - (max_count_xf * lengthscale_spacing_spanwise * np.cos(angle_orientation))
            * np.tan(angle_skew)
        )
        C_x = (
            min_count_yf * lengthscale_spacing_streamwise * np.sin(angle_orientation)
            + min_count_xf * lengthscale_spacing_spanwise * np.cos(angle_orientation)
            - (min_count_xf * lengthscale_spacing_spanwise * np.sin(angle_orientation))
            * np.tan(angle_skew)
        )
        C_y = (
            min_count_yf * lengthscale_spacing_streamwise * np.cos(angle_orientation)
            - min_count_xf * lengthscale_spacing_spanwise * np.sin(angle_orientation)
            - (min_count_xf * lengthscale_spacing_spanwise * np.cos(angle_orientation))
            * np.tan(angle_skew)
        )
        D_x = (
            -min_count_yf * lengthscale_spacing_streamwise * np.sin(angle_orientation)
            + min_count_xf * lengthscale_spacing_spanwise * np.cos(angle_orientation)
            - (min_count_xf * lengthscale_spacing_spanwise * np.sin(angle_orientation))
            * np.tan(angle_skew)
        )
        D_y = (
            -min_count_yf * lengthscale_spacing_streamwise * np.cos(angle_orientation)
            - min_count_xf * lengthscale_spacing_spanwise * np.sin(angle_orientation)
            - (min_count_xf * lengthscale_spacing_spanwise * np.cos(angle_orientation))
            * np.tan(angle_skew)
        )

        length_enclosing_farm_x = np.max([A_x, B_x, C_x, D_x]) - np.min(
            [A_x, B_x, C_x, D_x]
        )
        length_enclosing_farm_y = np.max([A_y, B_y, C_y, D_y]) - np.min(
            [A_y, B_y, C_y, D_y]
        )

        area_enclosingsquare_compass = (
            length_enclosing_farm_x + 2 * lengthscale_layback
        ) * (length_enclosing_farm_y + 2 * lengthscale_layback)
        outputs["area_tight"] = area_parallelopiped / (1e3) ** 2
        outputs["area_aligned_parcel"] = area_enclosingsquare_farmoriented / (1e3) ** 2
        outputs["area_compass_parcel"] = area_enclosingsquare_compass / (1e3) ** 2



import numpy as np
import jax.numpy as jnp
import jax
from ard.utils.mathematics import smooth_max, smooth_min, smooth_norm


def get_nearest_polygons(
    boundary_vertices,
    points_x,
    points_y,
    s=700,
    tol=1e-6,
):
    """
    Determines the nearest polygon for each point using the ray-casting algorithm. This
    function may be used to assign turbines to regions in a wind farm layout, but is not
    intended for use in a gradient-based optimization context. The function is not
    differentiable. This implementation is based on FLOWFarm.jl
    (https://github.com/byuflowlab/FLOWFarm.jl)

    Args:
        boundary_vertices (np.ndarray or list of np.ndarray): Vertices of the boundary in
            counterclockwise order. If `discrete` is True, this should be a list of arrays.
        points_x (np.ndarray): points x coordinates.
        points_y (np.ndarray): points y coordinates.
        s (float, optional): Smoothing factor for smooth max. Defaults to 700.
        tol (float, optional): Tolerance for determining proximity. Defaults to 1e-6.

    Returns:
        np.ndarray: Region assignments for each turbine.
    """

    # Number of turbines
    n_points = len(points_x)

    # Number of regions
    nregions = len(boundary_vertices)

    # Initialize region and status arrays
    region = -np.ones(n_points, dtype=int)

    # Initialize array to hold distances from each turbine to the closest boundary face
    turbine_to_region_distance = np.zeros(nregions, dtype=float)

    for i in range(n_points):
        # Iterate through each region
        for k in range(nregions):
            # Get distance to each region
            ctmp = distance_point_to_polygon_ray_casting(
                np.array([points_x[i], points_y[i]]),
                boundary_vertices[k],
                s=s,
                shift=tol,
                return_distance=True,
            )
            if ctmp <= 0:  # Negative if in boundary
                region[i] = k
                break

            # Check if the point is not in any of the regions
            else:
                # Calculate distance to each region
                turbine_to_region_distance[k] = ctmp

        # Magnitude of the constraint value
        if region[i] == -1:
            # Indicate closest region
            region[i] = np.argmin(turbine_to_region_distance)

    return region


def distance_multi_point_to_multi_polygon_ray_casting(
    points_x: np.ndarray[float],
    points_y: np.ndarray[float],
    boundary_vertices: list[list[np.ndarray]],
    regions: np.ndarray[int],
    s=700,
    tol=1e-6,
) -> np.ndarray:
    """
    Calculate the distance from each point to the nearest point on a polygon or set of polygons using
    the ray-casting (Jordan curve theorem) algorithm. Negative means the turbine is inside at least
    one polygon. This implementation is based on FLOWFarm.jl (https://github.com/byuflowlab/FLOWFarm.jl)

    Args:
        points_x (np.ndarray[list]): points x coordinates.
        points_y (np.ndarray[list]): points y coordinates.
        boundary_vertices (list[list[np.ndarray]]): Vertices of the each boundary in
            counterclockwise order. Boundaries should be simple polygons but do not need to have the same number of vertices.
        regions (np.array[int]): Predefined region assignments for each point. Defaults to None.
        s (float, optional): Smoothing factor for smooth max. Defaults to 700.
        tol (float, optional): Tolerance for determining proximity of point to polygon to be considered inside the polygon. Defaults to 1e-6.
        c (np.ndarray, optional): Preallocated array for constraint values. Defaults to None.

    Returns:
        np.ndarray: Constraint values for each turbine.
        np.ndarray (optional): Region assignments for each turbine (if `return_region` is True).
    """
    # Combine points_x and points_y into a single array of points
    points = jnp.stack([points_x, points_y], axis=1)

    # Convert boundary_vertices to JAX arrays
    boundary_vertices_jax = [
        jnp.asarray(poly, dtype=jnp.float32) for poly in boundary_vertices
    ]

    # Create a function for each polygon that computes distance
    def make_distance_func(vertices):
        def compute_for_polygon(point):
            return distance_point_to_polygon_ray_casting(
                point=point,
                vertices=vertices,
                s=s,
                shift=tol,
                return_distance=True,
            )

        return compute_for_polygon

    # Create list of functions, one per polygon
    distance_funcs = [
        make_distance_func(vertices) for vertices in boundary_vertices_jax
    ]

    # Define function to compute distance using jax.lax.switch
    def compute_distance(point, region_idx):
        return jax.lax.switch(region_idx, distance_funcs, point)

    # Vectorize over all points
    distances = jax.vmap(compute_distance)(points, regions)

    return distances


distance_multi_point_to_multi_polygon_ray_casting = jax.jit(
    distance_multi_point_to_multi_polygon_ray_casting
)


# Define a function to process a single edge
def process_edge(
    edge_start: jnp.ndarray,
    edge_end: jnp.ndarray,
    point: jnp.ndarray,
    shift: float,
):
    """
    Process a single polygon edge for ray-casting algorithm.

    Determines if a vertical ray cast from the point up intersects
    the edge, and calculates the distance from the point to the edge segment.

    Args:
        edge_start (jnp.ndarray): Start vertex of the edge ([x, y]).
        edge_end (jnp.ndarray): End vertex of the edge ([x, y]).
        point (jnp.ndarray): Point of interest ([x, y]).
        shift (float): Small value added to avoid division by zero when edge is vertical.

    Returns:
        tuple: A tuple containing:
            - is_below (bool): True if point is below the edge at the ray intersection.
            - distance (float): Shortest distance from point to the edge segment.
            - vertex_crossing (int): 1 if ray crosses exactly at edge_start vertex, 0 otherwise.
    """

    # Check if the x-coordinate of the point is between the x-coordinates of the edge,
    # including when the point is directly below a vertical edge
    x_condition = (
        ((edge_start[0] <= point[0]) & (point[0] < edge_end[0]))
        | ((edge_start[0] >= point[0]) & (point[0] > edge_end[0]))
        | (
            (
                jnp.isclose(edge_start[0], edge_end[0])
                & jnp.isclose(point[0], edge_start[0])
            )
        )
    )

    # Calculate the y-coordinate of the edge at the x-coordinate of the point
    y = ((edge_end[1] - edge_start[1]) / (edge_end[0] - edge_start[0] + shift)) * (
        point[0] - edge_start[0] + shift
    ) + edge_start[1]

    # Determine if the point is below the edge
    is_below = x_condition & (point[1] < y)

    # record vertex crossings
    vertex_crossing = jnp.where(
        x_condition & (edge_start[0] == point[0]) & is_below, 1, 0
    )

    # Calculate the distance to the edge
    distance = distance_point_to_lineseg_nd(point, edge_start, edge_end)

    return is_below, distance, vertex_crossing


# Vectorize the edge processing function
process_edge_vec = jax.vmap(process_edge, in_axes=(0, 0, None, None))


def distance_point_to_polygon_ray_casting(
    point: jnp.ndarray,
    vertices: jnp.ndarray,
    s: float = 700,
    shift: float = 1e-10,
    return_distance: bool = True,
):
    """
    Determines the signed distance from a point to a polygon using the Jordan curve
    theorem (ray-casting) approach as discussed in [1] and [2]. The polygon is
    assumed to be simple and defined in counterclockwise order. Complex polygons
    (where edges cross one another) are not supported. The function is
    differentiable with respect to the point coordinates.

    [1] Numerical Recipes: The Art of Scientific Computing by Press, et al. 3rd edition, sec. 21.4.3 (p. 1124)
    [2] https://en.wikipedia.org/wiki/Point_in_polygon

    Args:
        point (jnp.ndarray): Point of interest (2D vector).
        vertices (jnp.ndarray): Vertices of the polygon (Nx2 array) in counterclockwise order.
        s (float, optional): Smoothing factor for the smoothmin function. Defaults to 700.
        shift (float, optional): Small shift to handle edge cases. Defaults to 1e-10.
        return_distance (bool, optional): Whether to return the signed distance or just
            inside/outside status. Defaults to True. When False, the function is not
            differentiable.

    Returns:
        float: Signed distance or inside/outside status. Negative if inside, positive if outside.
    """

    # Ensure inputs are JAX arrays with explicit data types
    point = jnp.asarray(point, dtype=jnp.float32)
    vertices = jnp.asarray(vertices, dtype=jnp.float32)

    # Add the first vertex to the end to close the polygon loop
    vertices = jnp.vstack([vertices, vertices[0]])

    edge_starts = vertices[:-1]
    edge_ends = vertices[1:]
    is_below, distances, vertex_crossings = process_edge_vec(
        edge_starts, edge_ends, point, shift
    )

    # Check for tangential vertex crossings
    # Create arrays of current and previous edge differences
    curr_edge_end_diff = edge_ends[:, 0] - point[0]
    prev_edge_start_diff = (
        jnp.roll(edge_starts[:, 0], 1) - point[0]
    )  # Shift by 1 to get previous
    # Tangential crossing occurs when vertex_crossing > 0 AND the other two connected vertices are on same side
    tangential_vertex_crossings = jnp.where(
        (vertex_crossings > 0) & (curr_edge_end_diff * prev_edge_start_diff > 0), 1, 0
    )

    # Count the number of intersections, ignoring tangential vertex crossings
    intersection_counter = jnp.sum(is_below) - jnp.sum(tangential_vertex_crossings)

    # Compute the signed distance if desired
    if return_distance:
        c_prime = smooth_min(distances, s=s)
    else:
        c_prime = 1

    c = jax.lax.cond(
        intersection_counter % 2 == 1,
        lambda _: -c_prime,
        lambda _: c_prime,
        operand=None,
    )

    return c


def polygon_normals_calculator(
    boundary_vertices: np.ndarray, n_polygons: int = 1
) -> list[np.ndarray]:
    """
    Calculate unit vectors perpendicular to each edge of each polygon in a set of polygons.
    This implementation is based on FLOWFarm.jl (https://github.com/byuflowlab/FLOWFarm.jl).
    This function is not intended to be differentiable wrt n_polygons.

    Args:
        boundary_vertices (list of np.ndarray): List of m-by-2 arrays, where each array contains
            the boundary vertices of a polygon in counter-clockwise order. The number of vertices (m)
            in each polygon does not need to be the same.
        nboundaries (int, optional): The number of boundaries in the set. Defaults to 1.

    Returns:
        list of np.ndarray: List of m-by-2 arrays of unit vectors perpendicular to each edge
            of each polygon.
    """
    if n_polygons == 1:
        # Single polygon case
        return single_polygon_normals_calculator(boundary_vertices)
    else:
        # Multiple polygons case
        return multi_polygon_normals_calculator(boundary_vertices)


def multi_polygon_normals_calculator(
    boundary_vertices: list[np.ndarray],
) -> jnp.ndarray:
    """
    Calculate unit vectors perpendicular to each edge of each polygon in a set of polygons.
    This implementation is based on FLOWFarm.jl (https://github.com/byuflowlab/FLOWFarm.jl).

    Args:
        boundary_vertices (list of np.ndarray): List of m-by-2 arrays, where each array contains
            the boundary vertices of a polygon in counterclockwise order.
        nboundaries (int, optional): The number of boundaries in the set. Defaults to 1.

    Returns:
        list of np.ndarray: List of m-by-2 arrays of unit vectors perpendicular to each edge
            of each polygon.
    """
    return [
        single_polygon_normals_calculator(vertices) for vertices in boundary_vertices
    ]


multi_polygon_normals_calculator = jax.jit(multi_polygon_normals_calculator)


def single_polygon_normals_calculator(boundary_vertices: np.ndarray) -> jnp.ndarray:
    """
    Calculate unit vectors perpendicular to each edge of a polygon pointing into the polygon.
    This implementation is based on FLOWFarm.jl (https://github.com/byuflowlab/FLOWFarm.jl).

    Args:
        boundary_vertices (np.ndarray): m-by-2 array containing all the boundary vertices
            in counterclockwise order.

    Returns:
        np.ndarray: m-by-2 array of unit vectors perpendicular to each edge of the polygon pointing into the polygon.
    """

    # Add the first vertex to the end to form a closed loop
    boundary_vertices = jnp.vstack([boundary_vertices, boundary_vertices[0]])

    # Compute the differences (dx, dy) for each edge
    dx = boundary_vertices[1:, 0] - boundary_vertices[:-1, 0]
    dy = boundary_vertices[1:, 1] - boundary_vertices[:-1, 1]

    # Create vectors normal to the boundary edges
    boundary_normals = jnp.stack([-dy, dx], axis=1)

    # Normalize the vectors
    norms = jnp.linalg.norm(boundary_normals, axis=1, keepdims=True)
    boundary_normals = boundary_normals / norms

    return [boundary_normals]


single_polygon_normals_calculator = jax.jit(single_polygon_normals_calculator)


def point_on_line(p: np.ndarray, v1: np.ndarray, v2: np.ndarray, tol=1e-6):
    """
    Determine if a point lies on a line segment.

    Given a line determined by two points (v1 and v2), determine if the point (p) lies on the line
    between those points within a given tolerance.

    Args:
        p (np.ndarray): Point of interest (2D vector).
        v1 (np.ndarray): First vertex of the line (2D vector).
        v2 (np.ndarray): Second vertex of the line (2D vector).
        tol (float): Tolerance for determining co-linearity.

    Returns:
        bool: True if the point lies on the line, False otherwise.
    """

    d = distance_point_to_lineseg_nd(p, v1, v2)

    return jnp.isclose(d, 0.0, atol=tol)


def _distance_lineseg_to_lineseg_coplanar(
    line_a_start: np.ndarray,
    line_a_end: np.ndarray,
    line_b_start: np.ndarray,
    line_b_end: np.ndarray,
) -> float:
    """Returns the distance between two finite line segments assuming the segments are coplanar.
    It is up to the user to check the required condition. There may be some error in the case
    when the line segments are parallel since multiple points may have equal distances, leading to
    some error from the smooth minimum function.

    Args:
        line_a_start (np.ndarray): start point of line a
        line_a_end (np.ndarray): end point of line a
        line_b_start (np.ndarray): start point of line b
        line_b_end (np.ndarray): end point of line b

    Returns:
        distance (float): the distance between the lines
    """

    # get distance between all pairs of end points
    a_start_to_b = distance_point_to_lineseg_nd(line_a_start, line_b_start, line_b_end)
    a_end_to_b = distance_point_to_lineseg_nd(line_a_end, line_b_start, line_b_end)
    b_start_to_a = distance_point_to_lineseg_nd(line_b_start, line_a_start, line_a_end)
    b_end_to_a = distance_point_to_lineseg_nd(line_b_end, line_a_start, line_a_end)

    distance = smooth_min(
        jnp.array([a_start_to_b, a_end_to_b, b_start_to_a, b_end_to_a])
    )

    return distance


_distance_lineseg_to_lineseg_coplanar = jax.jit(_distance_lineseg_to_lineseg_coplanar)


def distance_lineseg_to_lineseg_nd(
    line_a_start: np.ndarray,
    line_a_end: np.ndarray,
    line_b_start: np.ndarray,
    line_b_end: np.ndarray,
    tol=1e-12,
) -> float:
    """Find the distance between two line segments in 2d or 3d. This method is primarily based on reference [1],
    using a parametric approach based on the determinant and cross product to find the closest points on the two line
    segments. However, to handle the special case of line segments that are coplanar, we use the smooth minimum of the
    distance between the endpoints of the two line segments and the other line segment. In the coplanar case, the
    returned distance between the two line segments may have a noticeable error due to possibly having multiple points
    with the same distance, which leads to error in the smooth minimum function.

    [1] Numerical Recipes: The Art of Scientific Computing by Press, et al. 3rd edition, sec. 21.4.2 (p. 1121)

    Args:
        line_a_start (np.ndarray): The start point of line segment "a" as either [x,y,z] or [x,y]
        line_a_end (np.ndarray): The end point of line segment "a" as either [x,y,z] or [x,y]
        line_b_start (np.ndarray): The start point of line segment "b" as either [x,y,z] or [x,y]
        line_b_end (np.ndarray): The end point of line segment "b" as either [x,y,z] or [x,y]
        tol (float, optional): If denominator in key equation is less than or equal tol, then an alternative method is used. Defaults to 0.0.

    Returns:
        float: Distance between the two line segments
    """

    def a_is_point(inputs0i) -> float:
        line_a_start = inputs0i[0]
        line_b_start = inputs0i[2]
        line_b_end = inputs0i[3]
        return distance_point_to_lineseg_nd(line_a_start, line_b_start, line_b_end)

    def b_is_point(inputs0i) -> float:
        line_a_start = inputs0i[0]
        line_a_end = inputs0i[1]
        line_b_start = inputs0i[2]
        return distance_point_to_lineseg_nd(line_b_start, line_a_start, line_a_end)

    def a_is_not_point(inputs0i) -> float:
        line_b_vector = inputs0i[5]
        return jax.lax.cond(
            jnp.all(line_b_vector == 0.0), b_is_point, a_and_b_are_lines, inputs0i
        )

    def a_and_b_are_lines(inputs0i) -> float:

        def denom_lt_tol(inputs1i) -> float:
            line_a_start = inputs1i[0]
            line_a_end = inputs1i[1]
            line_b_start = inputs1i[2]
            line_b_end = inputs1i[3]
            return _distance_lineseg_to_lineseg_coplanar(
                line_a_start=line_a_start,
                line_a_end=line_a_end,
                line_b_start=line_b_start,
                line_b_end=line_b_end,
            )

        def denom_gt_tol(inputs1i) -> float:
            line_a_start = inputs1i[0]
            line_a_end = inputs1i[1]
            line_b_start = inputs1i[2]
            line_b_end = inputs1i[3]
            line_a_vector = inputs1i[4]
            line_b_vector = inputs1i[5]
            denominator = inputs1i[6]

            a = line_a_start
            v = line_a_vector
            x = line_b_start
            u = line_b_vector

            s_numerator = jnp.linalg.det(jnp.array([a - x, u, jnp.cross(u, v)]).T)
            t_numerator = jnp.linalg.det(jnp.array([a - x, v, jnp.cross(u, v)]).T)

            s = s_numerator / denominator
            t = t_numerator / denominator

            # Get closest point along the lines
            # if s or t > 1, use end point of line
            def st_gt_1(inputs23i) -> np.ndarray:
                line_end = inputs23i[1]
                return jnp.array(line_end, dtype=float)

            # if s or t < 0, use start point of line
            def st_lt_0(inputs23i) -> np.ndarray:
                line_start = inputs23i[0]
                return jnp.array(line_start, dtype=float)

            # otherwise compute the closest point on line using the parametric form of the line segment
            def st_gt_0_lt_1(inputs23i) -> np.ndarray:
                line_start = inputs23i[0]
                line_vector = inputs23i[2]
                st = inputs23i[3]
                return jnp.array(line_start + st * line_vector, dtype=float)

            def st_lt_1(inputs23i) -> np.ndarray:
                st = inputs23i[3]
                return jax.lax.cond(st < 0, st_lt_0, st_gt_0_lt_1, inputs23i)

            # get closest point on lines a and b to each other
            inputs2o = [line_a_start, line_a_end, line_a_vector, s]
            closest_point_line_a = jax.lax.cond(s > 1, st_gt_1, st_lt_1, inputs2o)
            inputs3o = [line_b_start, line_b_end, line_b_vector, t]
            closest_point_line_b = jax.lax.cond(t > 1, st_gt_1, st_lt_1, inputs3o)

            # the distance between the line segments is the distance between the closest points (in many cases)
            parametric_distance = smooth_norm(
                closest_point_line_b - closest_point_line_a
            )

            # parametric approach can miss cases, so compare with point to line distances
            distance_point_a_line_b = distance_point_to_lineseg_nd(
                closest_point_line_a, line_b_start, line_b_end
            )
            distance_point_b_line_a = distance_point_to_lineseg_nd(
                closest_point_line_b, line_a_start, line_a_end
            )
            distance = smooth_min(
                jnp.array(
                    [
                        parametric_distance,
                        distance_point_a_line_b,
                        distance_point_b_line_a,
                    ]
                )
            )

            return distance

        line_a_start = inputs0i[0]
        line_a_end = inputs0i[1]
        line_b_start = inputs0i[2]
        line_b_end = inputs0i[3]
        line_a_vector = inputs0i[4]
        line_b_vector = inputs0i[5]

        # find s and t (point along segment where the segments are closest to each other) using eq. 21.4.17 in [1]
        denominator = smooth_norm(jnp.cross(line_b_vector, line_a_vector)) ** 2

        inputs1o = [
            line_a_start,
            line_a_end,
            line_b_start,
            line_b_end,
            line_a_vector,
            line_b_vector,
            denominator,
        ]

        distance = jax.lax.cond(
            denominator <= tol, denom_lt_tol, denom_gt_tol, inputs1o
        )

        return distance

    # if 2d given, then pad with zeros to get 3d points
    pad_width = len(line_a_start)
    line_a_start = jnp.pad(line_a_start, (0, 3 - pad_width))
    line_a_end = jnp.pad(line_a_end, (0, 3 - pad_width))
    line_b_start = jnp.pad(line_b_start, (0, 3 - pad_width))
    line_b_end = jnp.pad(line_b_end, (0, 3 - pad_width))

    line_a_vector = line_a_end - line_a_start
    line_b_vector = line_b_end - line_b_start

    inputs0o = [
        line_a_start,
        line_a_end,
        line_b_start,
        line_b_end,
        line_a_vector,
        line_b_vector,
    ]

    distance = jax.lax.cond(
        jnp.all(line_a_vector == 0.0), a_is_point, a_is_not_point, inputs0o
    )

    return distance


distance_lineseg_to_lineseg_nd = jax.jit(distance_lineseg_to_lineseg_nd)


def distance_point_to_lineseg_nd(
    point: np.ndarray, segment_start: np.ndarray, segment_end: np.ndarray
) -> float:
    """Find the distance from a point to a line segment in N-Dimensions. This
    implementation can handle any number of dimensions as well as the reduced case
    of point to point distance. If the same point is passed for the start and end
    of the line segment, then the distance is simply the distance from the point
    of interest to the single start/end point.

    Args:
        point (np.ndarray): point of interest [x,y,...]
        segment_start (np.ndarray): start point of line segment [x,y,...]
        segment_end (np.ndarray): end point of line segment [x,y,...]

    Returns:
        distance (float): shortest distance between the point and finite line segment
    """

    def if_point_to_point(inputs) -> float:
        point = inputs[0]
        segment_start = inputs[1]
        return jnp.float64(smooth_norm(segment_start - point))

    def if_point_to_line_seg(inputs) -> float:
        point = inputs[0]
        segment_start = inputs[1]
        segment_end = inputs[2]
        segment_vector = inputs[3]

        # get the closest point on the line segment to the point of interest
        closest_point = get_closest_point_on_line_seg(
            point, segment_start, segment_end, segment_vector
        )

        # the distance from the point to the line is the distance from the point to the closest point on the line
        return jnp.float64(smooth_norm(point - closest_point))

    # get the vector of the line segment
    segment_vector = segment_end - segment_start

    # if the segment is a point, then get the distance to that point
    distance = jax.lax.cond(
        jnp.all(segment_vector == 0),
        if_point_to_point,
        if_point_to_line_seg,
        [point, segment_start, segment_end, segment_vector],
    )

    return distance


distance_point_to_lineseg_nd = jax.jit(distance_point_to_lineseg_nd)


def get_closest_point_on_line_seg(
    point: np.ndarray,
    segment_start: np.ndarray,
    segment_end: np.ndarray,
    segment_vector: np.ndarray,
) -> np.ndarray:
    """Get the closest point on a line segment to the point of interest in N-Dimensions
    using vector projection.

    Args:
        point (np.ndarray): point of interest [x,y,...]
        segment_start (np.ndarray): start point of line segment [x,y,...]
        segment_end (np.ndarray): end point of line segment [x,y,...]
        segment_vector (np.ndarray): segment_end - segment_start

    Returns:
        np.ndarray: closest point on the line segment to the point of interest
    """

    # calculate the distance to the starting point
    start_to_point_vector = point - segment_start

    # calculate the unit vector projection of the start to point vector on the line segment
    projection = jnp.divide(
        jnp.dot(start_to_point_vector, segment_vector),
        jnp.dot(segment_vector, segment_vector),
    )

    def lt_0(inputs) -> np.ndarray:
        segment_start = inputs[1]
        return jnp.array(segment_start, dtype=jnp.float64)

    def gt_1(inputs) -> np.ndarray:
        segment_end = inputs[2]
        return jnp.array(segment_end, dtype=jnp.float64)

    def gt_0(inputs) -> np.ndarray:
        projection = inputs[0]
        return jax.lax.cond(projection > 1, gt_1, lt_1_gt_0, inputs)

    def lt_1_gt_0(inputs) -> np.ndarray:
        projection = inputs[0]
        segment_start = inputs[1]
        segment_vector = inputs[3]
        return jnp.array(segment_start + projection * segment_vector, dtype=jnp.float64)

    return jax.lax.cond(
        projection < 0,
        lt_0,
        gt_0,
        [projection, segment_start, segment_end, segment_vector],
    )


get_closest_point_on_line_seg = jax.jit(get_closest_point_on_line_seg)



from __future__ import annotations  # for type hinting annotation fix...

import numpy as np

from floris.wind_data import WindDataBase
from floris.wind_data import TimeSeries


class WindQuery:
    """
    A class that manages sets of wind conditions to query that can operate
    for different types and fidelities of aerodynamic solvers without loss of
    generality.
    """

    def __init__(
        self,
        directions: np.ndarray = None,
        speeds: np.ndarray = None,
        TIs: np.ndarray = None,
    ) -> None:
        """
        WindQuery initialization.

        A WindQuery object must be initialized with speeds and directions, which
        must be equal-length numpy arrays, and turbulence intensities can be
        provided with a same-length array, a single float, or will otherwise be
        set by default using FLORIS's `assign_ti_using_IEC_method`.

        Parameters
        ----------
        directions : np.ndarray
            the directions of the wind resource to be evaluated in degrees
        speeds : np.ndarray
            the velocities of the wind resource to be evaluated in meters/second
        TIs : np.ndarray, optional
            turbulence intensities of the wind resource (non-dimensional)
        """

        self.directions = np.array([])
        self.speeds = np.array([])
        self.TIs = np.array([])

        if directions is not None:
            self.set_directions(directions)
        if speeds is not None:
            self.set_speeds(speeds)
        if TIs is not None:
            self.set_TIs(TIs)

    def set_directions(self, directions: np.ndarray) -> None:
        """
        Set the directions on a WindQuery object.

        Parameters
        ----------
        directions : np.ndarray
            the directions of the wind resource to be assigned, in degrees
        """

        self.directions = directions
        self.N_conditions = (
            None
            if (
                (self.directions.size != self.speeds.size)
                or (self.directions.size != self.TIs.size)
            )
            else self.directions.size
        )

    def set_speeds(self, speeds: np.ndarray) -> None:
        """
        Set the wind speeds on a WindQuery object.

        Parameters
        ----------
        speeds : np.ndarray
            the speeds of the wind resource to be assigned, in meters/second
        """

        self.speeds = speeds
        self.N_conditions = (
            None
            if (
                (self.directions.size != self.speeds.size)
                or (self.directions.size != self.TIs.size)
            )
            else self.directions.size
        )

    def set_TIs(self, TIs: float | np.ndarray) -> None:
        """
        Set the turbulence intensities on a WindQuery object.

        Parameters
        ----------
        TIs : float or np.ndarray
            the turbulence intensity value or values of the wind resource to be
            assigned, non-dimensionally
        """

        # flip over to using a numpy array and correct size if set as a float
        TIs = np.array(TIs)
        if (np.array(TIs).size == 1) and (self.N_conditions is not None):
            TIs = TIs * np.ones((self.N_conditions,))
        elif np.array(TIs).size == 1:
            assert np.all(
                self.directions.shape == self.speeds.shape
            ), "to set TIs automatically direction and speed must be set, and consistently"
            TIs = TIs * np.ones_like(self.directions)
        else:
            assert np.all(TIs.shape == self.directions.shape) and np.all(
                TIs.shape == self.speeds.shape
            ), "mismatch in TI size vs. direction"
            assert np.all(
                TIs.shape == self.speeds.shape
            ), "mismatch in TI size vs. speed"
        self.TIs = TIs
        self.N_conditions = (
            None
            if (
                (self.directions.size != self.speeds.size)
                or (self.directions.size != self.TIs.size)
            )
            else self.directions.size
        )

    def set_TI_using_IEC_method(self) -> None:
        """
        Re-set the turbulence intensities using the FLORIS IEC method interface.
        """

        assert self.directions.size != 0, "directions must be set"
        assert self.speeds.size != 0, "speeds must be set"
        # use a temporary FLORIS time series to get the IEC TIs
        ts_temp = TimeSeries(
            wind_directions=self.directions,
            wind_speeds=self.speeds,
            turbulence_intensities=0.06
            * np.ones_like(self.directions),  # default value
        )
        ts_temp.assign_ti_using_IEC_method()

        # re-set all the values
        _, _, TIs_new, _, _, _ = ts_temp.unpack()
        self.set_TIs(TIs_new)

    def get_directions(self) -> np.ndarray:
        """Get the directions from the wind query object."""
        assert self.is_valid(), "mismatch in direction/speed vectors"
        return self.directions

    def get_speeds(self) -> np.ndarray:
        """Get the wind speeds from the wind query object."""
        assert self.is_valid(), "mismatch in direction/speed vectors"
        return self.speeds

    def get_TIs(self) -> np.ndarray:
        """Get the turbulence intensities from the wind query object."""
        assert self.is_valid(), "mismatch in direction/speed vectors"
        return self.TIs

    def is_valid(self) -> bool:
        """Ensure that the specified wind conditions are valid."""

        # first, not a valid query if the directions and speeds aren't specified
        if (self.directions.size == 0) or (self.speeds.size == 0):
            return False
        # next, to be valid the directions and speeds should be the same size and shape
        if not np.all(np.equal(self.directions.shape, self.speeds.shape)):
            return False
        # next, to be valid the TIs should also be the same size and shape. or empty
        if (not np.all(np.equal(self.directions.shape, self.TIs.shape))) and (
            self.TIs.shape != ()
        ):
            return False
        # to be valid, directions should be on [0, 360]
        if np.any((self.directions < 0.0) | (self.directions > 360.0)):
            return False
        # to be valid, velocity should be on [0, +inf)
        if np.any(self.speeds < 0.0):
            return False
        # to be valid, TI should be on [0, +inf)
        if np.any(self.TIs < 0.0):
            return False
        return True

    def from_FLORIS_WindData(winddata_FLORIS: WindDataBase) -> WindQuery:
        """
        Turn a FLORIS WindData object into a (more general) WindQuery object.

        Parameters
        ----------
        winddata_FLORIS : floris.wind_data.WindDataBase
            A FLORIS wind data object derived from the WindDataBase base class.

        Returns
        -------
        WindQuery
            A WindQuery object that represents that same wind data as the FLORIS
            wind data object.
        """

        wind_directions, wind_speeds, ti_table, _, _, _ = winddata_FLORIS.unpack()

        # create the wind query for this condition
        wq = WindQuery(wind_directions, wind_speeds, TIs=ti_table)
        assert wq.is_valid()  # make sure it's legit
        return wq  # and ship it



import openmdao.api as om


class SpacingApproximations(om.ExplicitComponent):
    """
    OpenMDAO component to calculate approximations for turbine spacing based on the total length of cables
    and the number of wind turbines.

    Inputs
    ------
    total_length_cables : float
        Total length of cables in meters.

    Outputs
    -------
    primary_turbine_spacing_diameters : float
        Approximation of spacing between turbines in diameters for use in cost estimation using LandBOSSE.
    secondary_spacing_diameters : float
        Approximation of spacing between rows of turbines in diameters for use in cost estimation using LandBOSSE.

    Options
    -------
    modeling_options : dict
        Dictionary of modeling options including at least ["farm"]["N_turbines"] and ["turbine"]["geometry"]["diameter_rotor"]
    """

    def initialize(self):
        """Initialize the component and declare options."""
        self.options.declare(
            "modeling_options", types=dict, desc="Ard modeling options"
        )

    def setup(self):
        """Set up the inputs and outputs."""
        self.add_input(
            "total_length_cables", val=0.0, units="m", desc="Total cable length"
        )
        self.add_output(
            "primary_turbine_spacing_diameters",
            val=0.0,
            units=None,
            desc="Turbine spacing",
        )
        self.add_output(
            "secondary_turbine_spacing_diameters",
            val=0.0,
            units=None,
            desc="Row spacing",
        )

    def setup_partials(self):
        """Declare partial derivatives."""
        N_turbines = self.options["modeling_options"]["layout"]["N_turbines"]
        rotor_diameter_m = self.options["modeling_options"]["windIO_plant"][
            "wind_farm"
        ]["turbine"]["rotor_diameter"]

        # Partial derivative of primary_turbine_spacing_diameters w.r.t. total_length_cables are constant
        const_partial = 1.0 / (rotor_diameter_m * N_turbines)
        self.declare_partials(
            "primary_turbine_spacing_diameters",
            "total_length_cables",
            val=const_partial,
        )
        self.declare_partials(
            "secondary_turbine_spacing_diameters",
            "total_length_cables",
            val=const_partial,
        )

    def compute(self, inputs, outputs):
        """Compute the turbine spacing."""
        total_length_cables = inputs["total_length_cables"]
        N_turbines = self.options["modeling_options"]["layout"]["N_turbines"]
        rotor_diameter_m = self.options["modeling_options"]["windIO_plant"][
            "wind_farm"
        ]["turbine"]["rotor_diameter"]

        # Calculate turbine and row spacing
        outputs["primary_turbine_spacing_diameters"] = total_length_cables / (
            rotor_diameter_m * N_turbines
        )
        outputs["secondary_turbine_spacing_diameters"] = total_length_cables / (
            rotor_diameter_m * N_turbines
        )

    def compute_partials(self, inputs, partials):
        "partials are constant, so no calculations needed here"
        pass



import numpy as np
import openmdao.api as om

from shapely import length
import shapely.geometry as sg

import ard.layout.templates


class FullFarmLanduse(ard.layout.templates.LanduseTemplate):
    """
    Landuse class for full Cartesian grid farm layout.

    This is a class to compute the landuse area of a fully specified Cartesian
    grid farm layout.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary (inherited from
        `templates.LanduseTemplate`)

    Inputs
    ------
    x_turbines : np.ndarray
        a 1-D numpy array that represents that x (i.e. Easting) coordinate of
        the location of each of the turbines in the farm in meters
    y_turbines : np.ndarray
        a 1-D numpy array that represents that y (i.e. Northing) coordinate of
        the location of each of the turbines in the farm in meters

    Outputs
    -------
    area_tight : float
        the area in square kilometers that the farm occupies based on the
        circumscribing geometry with a specified (default zero) layback buffer
        (inherited from `templates.LayoutTemplate`)
    """

    def setup(self):
        """Setup of OM component."""
        super().setup()

        # add the full layout inputs
        self.add_input(
            "x_turbines",
            np.zeros((self.N_turbines,)),
            units="m",
            desc="turbine location in x-direction",
        )
        self.add_input(
            "y_turbines",
            np.zeros((self.N_turbines,)),
            units="m",
            desc="turbine location in y-direction",
        )

    def setup_partials(self):
        """Derivative setup for OM component."""

        # default complex step for the layout-landuse tools, since they're often algebraic
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        """Computation for the OM component."""

        # extract the points from the inputs
        points = list(
            zip(
                list(inputs["x_turbines"]),
                list(inputs["y_turbines"]),
            )
        )

        # create a multi-point object
        mp = sg.MultiPoint(points)

        # create a laybacked geometry
        D_rotor = self.windIO["wind_farm"]["turbine"]["rotor_diameter"]
        lengthscale_layback = float(inputs["distance_layback_diameters"][0] * D_rotor)

        # area tight is equal to the convex hull area for the points in sq. km.
        outputs["area_tight"] = (
            mp.convex_hull.buffer(lengthscale_layback).area / 1000**2
        )



import importlib
import openmdao.api as om
from openmdao.drivers.doe_driver import DOEGenerator
from wisdem.optimization_drivers.nsga2_driver import NSGA2Driver
from openmdao.utils.file_utils import clean_outputs
from ard.utils.io import load_yaml, replace_key_value
from ard.utils.logging import prepend_tabs_to_stdio
from ard.cost.wisdem_wrap import (
    LandBOSSE_setup_latents,
    ORBIT_setup_latents,
    FinanceSE_setup_latents,
)
import windIO
from ard import ASSET_DIR
from typing import Union


def set_up_ard_model(input_dict: Union[str, dict], root_data_path: str = None):
    """
    Set up an OpenMDAO model for Ard based on the provided input dictionary or YAML file.

    This function initializes and configures an OpenMDAO problem using a system
    specification, modeling options, and analysis options. It supports default
    system configurations (e.g., "onshore", "offshore_floating") and allows for
    recursive setup of subsystems and connections.

    Parameters
    ----------
    input_dict : Union[str, dict]
        A dictionary or a path to a YAML file containing the configuration for the Ard model.
        The dictionary or YAML file must include:

        - "system" : str or dict
            The name of the default system to use (e.g., "onshore") or a custom system specification.
        - "modeling_options" : dict
            A dictionary defining the modeling options for the system (e.g., turbine specs, farm layout).
        - "analysis_options" : dict
            A dictionary defining the analysis options, including driver settings, design variables,
            constraints, objectives, and recorder configuration.

    root_data_path : str, optional
        The root path for resolving relative paths in the system configuration. Defaults to None.

    Returns
    -------
    om.Problem
        An OpenMDAO problem instance with the defined system hierarchy, modeling options,
        and analysis options.

    Raises
    ------
    ValueError
        If an invalid default system is specified or if required keys are missing in the input dictionary.

    Notes
    -----
    - The function uses `set_up_system_recursive` to recursively build the system hierarchy.
    - Latent variables for LandBOSSE, ORBIT, and FinanceSE are automatically set up if their
      respective components are present in the model.

    """

    # load dictionary if string is given
    if isinstance(input_dict, str):
        input_dict, root_data_path = load_yaml(input_dict, return_path=True)

    # load default system if requested and available
    available_default_systems = [
        "onshore",
        "onshore_batch",
        "onshore_no_cable_design",
        "offshore_monopile",
        "offshore_monopile_no_cable_design",
        "offshore_floating",
        "offshore_floating_no_cable_design",
    ]

    if isinstance(input_dict["system"], str):
        if input_dict["system"] in available_default_systems:
            system = load_yaml(ASSET_DIR / f"ard_system_{input_dict['system']}.yaml")

            input_dict["system"] = replace_key_value(
                target_dict=system,
                target_key="modeling_options",
                new_value=input_dict["modeling_options"],
            )
        else:
            raise (
                ValueError(
                    f"invalid default system '{input_dict['system']}' specified. Must be one of {available_default_systems}"
                )
            )

    # replace empty data_path specs
    input_dict["system"] = replace_key_value(
        target_dict=input_dict["system"],
        target_key="data_path",
        new_value=root_data_path,
        replace_none_only=True,
    )

    # validate windIO dictionary
    windIO_dict = input_dict["modeling_options"]["windIO_plant"]
    windIO.validate(windIO_dict, schema_type="plant/wind_energy_system")

    # set up the openmdao problem
    prob = set_up_system_recursive(
        input_dict=input_dict["system"],
        modeling_options=input_dict["modeling_options"],
        analysis_options=input_dict["analysis_options"],
    )

    return prob


def set_up_system_recursive(
    input_dict: dict,
    system_name: str = "top_level",
    case_name: str | None = None,
    work_dir: str = "case_files",
    parent_group=None,
    modeling_options: dict = None,
    analysis_options: dict = None,
    _depth: int = 0,
):
    """
    Recursively sets up an OpenMDAO system based on the input dictionary.

    Args:
        input_dict (dict): Dictionary defining the system hierarchy.
        parent_group (om.Group, optional): The parent group to which subsystems are added.
                                           Defaults to None, which initializes the top-level problem.

    Returns:
        om.Problem: The OpenMDAO problem with the defined system hierarchy.
    """

    # grab the case name if it's supplied in the system yaml
    if case_name is None:
        case_name = modeling_options.get(
            "case_name",
            input_dict.get("modeling_options", {}).get(
                "case_name",
                "ard_problem",
            ),
        )

    # Initialize the top-level problem if no parent group is provided
    if parent_group is None:
        # clean out any pre-existing results for this problem
        print("Running OpenMDAO util to clean the output directories...")
        prepend_tabs_to_stdio(clean_outputs)(recurse=True, prompt=False)
        print("... done.\n")

        prob = om.Problem(
            name=case_name,
            work_dir=work_dir,
        )
        parent_group = prob.model

        print(f"Created top-level OpenMDAO problem: {system_name}.")
    else:
        prob = None

    # Add subsystems directly from the input dictionary
    if hasattr(parent_group, "name") and (parent_group.name != ""):
        print(
            f"{''.join(' ' * 4 * _depth)}Adding {system_name} to {parent_group.name}."
        )
    else:
        print(f"{''.join(' ' * 4 * _depth)}Adding {system_name}.")
    if "systems" in input_dict:  # Recursively add nested subsystems]
        if _depth > 0:
            group = parent_group.add_subsystem(
                name=system_name,
                subsys=om.Group(),
                promotes=input_dict.get("promotes", None),
            )
        else:
            group = parent_group
        for subsystem_key, subsystem_data in input_dict["systems"].items():
            set_up_system_recursive(
                subsystem_data,
                parent_group=group,
                system_name=subsystem_key,
                modeling_options=modeling_options,
                analysis_options=None,
                _depth=_depth + 1,
            )
        if "approx_totals" in input_dict:
            prefix = "\t"
            print(prefix + f"Activating approximate totals on {system_name}")
            group.approx_totals(**input_dict["approx_totals"])

    else:
        subsystem_data = input_dict

        if "object" not in subsystem_data:
            raise ValueError(f"Ard subsystem '{system_name}' missing 'object' spec.")
        if "promotes" not in subsystem_data:
            raise ValueError(f"Ard subsystem '{system_name}' missing 'promotes' spec.")

        # Dynamically import the module and get the subsystem class
        Module = importlib.import_module(subsystem_data["module"])
        SubSystem = getattr(Module, subsystem_data["object"])

        # Convert specific promotes to tuples
        promotes = [
            tuple(p) if isinstance(p, list) else p for p in subsystem_data["promotes"]
        ]

        # Add the subsystem to the parent group with kwargs
        parent_group.add_subsystem(
            name=system_name,
            subsys=SubSystem(**subsystem_data.get("kwargs", {})),
            promotes=promotes,
        )

    # Handle connections within the parent group
    if "connections" in input_dict:
        for connection in input_dict["connections"]:
            src, tgt = connection  # Unpack the connection as [src, tgt]
            parent_group.connect(src, tgt)

    if _depth == 0:
        print(f"System {system_name} built.")

    # Set up the problem if this is the top-level call
    if prob is not None:

        if analysis_options:
            # set up driver
            if "driver" in analysis_options:

                name_driver = analysis_options["driver"]["name"]

                if name_driver == "NSGA2":
                    prob.driver = NSGA2Driver()
                else:
                    Driver = getattr(om, name_driver)

                    # handle DOE drivers with special treatment
                    if Driver == om.DOEDriver:
                        generator = None
                        if "generator" in analysis_options["driver"]:
                            if type(analysis_options["driver"]["generator"]) == dict:
                                gen_dict = analysis_options["driver"]["generator"]
                                generator = getattr(om, gen_dict["name"])(
                                    **gen_dict["args"]
                                )
                            elif isinstance(
                                analysis_options["driver"]["generator"], DOEGenerator
                            ):
                                generator = analysis_options["driver"]["generator"]
                            else:
                                raise NotImplementedError(
                                    "Only dictionary-specified or OpenMDAO "
                                    "DOEGenerator generators have been implemented."
                                )
                        prob.driver = Driver(generator)
                    else:
                        prob.driver = Driver()

                # handle the options now
                if "options" in analysis_options["driver"]:
                    for option, value_driver_option in analysis_options["driver"][
                        "options"
                    ].items():
                        if option == "opt_settings":
                            for (
                                key_opt_setting,
                                value_opt_setting,
                            ) in value_driver_option.items():
                                prob.driver.opt_settings[key_opt_setting] = (
                                    value_opt_setting
                                )
                        else:
                            prob.driver.options[option] = value_driver_option

                    # set design variables
            if "design_variables" in analysis_options:
                for var_name, var_data in analysis_options["design_variables"].items():
                    prob.model.add_design_var(var_name, **var_data)

            # set constraints
            if "constraints" in analysis_options:
                for constraint_name, constraint_data in analysis_options[
                    "constraints"
                ].items():
                    prob.model.add_constraint(constraint_name, **constraint_data)

            # set objective
            if "objectives" in analysis_options:
                for obj_name, obj_options in analysis_options["objectives"].items():
                    obj_options = {} if (obj_options is None) else obj_options
                    prob.model.add_objective(
                        obj_name,
                        **obj_options,
                    )

            # Set up the recorder if specified in the input dictionary
            if "recorder" in analysis_options:
                recorder_filepath = analysis_options["recorder"].get("filepath")
                if recorder_filepath:
                    recorder = om.SqliteRecorder(recorder_filepath)
                    prob.add_recorder(recorder)
                    prob.driver.add_recorder(recorder)

        # TODO! THIS IS NECESSARY FOR SOME REASON WHEN RUNNING FREE
        # OPTIMIZATIONS. THIS SHOULDN'T BE NEEDED...
        prob.model.set_input_defaults(
            "x_turbines",
            # input_dict["modeling_options"]["windIO_plant"]["wind_farm"]["layouts"]["coordinates"]["x"],
            units="m",
        )
        prob.model.set_input_defaults(
            "y_turbines",
            # input_dict["modeling_options"]["windIO_plant"]["wind_farm"]["layouts"]["coordinates"]["y"],
            units="m",
        )

        # setup the openmdao problem
        print(f"System {system_name} set up.")
        prob.setup()

    return prob



from pathlib import Path
from os import PathLike

import os
import yaml


class Loader(yaml.SafeLoader):

    def __init__(self, stream):

        self._root = os.path.split(stream.name)[0]

        super().__init__(stream)

    def include(self, node):

        filename = os.path.join(self._root, self.construct_scalar(node))

        with open(filename, "r") as f:
            return yaml.load(f, self.__class__)


Loader.add_constructor("!include", Loader.include)


def load_yaml(filename, loader=Loader, return_path=False) -> dict:
    if isinstance(filename, dict):
        return filename  # filename already yaml dict
    with open(filename) as fid:
        if return_path:
            return yaml.load(fid, loader), Path(filename).parent.absolute()
        else:
            return yaml.load(fid, loader)


def check_create_folder(filepath):
    already_exists = True
    if not os.path.isdir(filepath):
        os.makedirs(filepath, exist_ok=True)
        already_exists = False
    return already_exists


def write_yaml(filename, data):
    if not ".yaml" in filename:
        filename = filename + ".yaml"

    with open(filename, "w+") as file:
        yaml.dump(data, file, sort_keys=False, encoding=None, default_flow_style=False)


def replace_key_value(
    target_dict: dict, target_key: str, new_value, replace_none_only=True
) -> dict:
    """
    Recursively replace the value of a target key in a dictionary.

    Parameters:
    - target_dict (dict): The dictionary to process.
    - target_key (str): The key whose value needs to be replaced.
    - new_value: The new value to assign to the target key.
    - replace_none_only (bool): if True, only 'None' values will be replaced.

    Returns:
    - dict: The updated dictionary.
    """
    for key, value in target_dict.items():
        if key == target_key:
            if (value != None) and (replace_none_only):
                continue
            target_dict[key] = new_value
        elif isinstance(value, dict):
            # Recurse into nested dictionaries
            replace_key_value(
                value, target_key, new_value, replace_none_only=replace_none_only
            )
        elif isinstance(value, list):
            # Handle lists of dictionaries
            for item in value:
                if isinstance(item, dict):
                    replace_key_value(
                        item, target_key, new_value, replace_none_only=replace_none_only
                    )
    return target_dict



import numpy as np
import scipy.spatial

import ard.layout.templates as templates
import ard.layout.fullfarm as fullfarm

phi_golden = (1 + np.sqrt(5)) / 2  # golden ratio


def sunflower(
    n: float,
    alpha: float = 0,  # proportion of points that should end on boundary
    n_b: float = None,  # for overriding with the number of boundary elements
    geodesic=False,  # use geodesic step function
):
    """
    generate a sunflower seed packing pattern

    adapted from a stackoverflow post:
        https://stackoverflow.com/questions/28567166/uniformly-distribute-x-points-inside-a-circle#28572551

    in turn from the wolfram demonstrations page:
        Joost de Jong (2013),
        "Sunflower Seed Arrangements"
        Wolfram Demonstrations Project.
        demonstrations.wolfram.com/SunflowerSeedArrangements/.

    appears to originate from: doi:10.1016/0025-5564(79)90080-4
    """

    def radius(k: int, n: int, b: int):
        """
        radius at which a seed should live
        b sets the number of boundary points
        remainder of n points have sequence-baased location
        """

        if k > n - b:
            return 1.0
        else:
            return np.sqrt(k - 0.5) / np.sqrt(n - (b + 1) / 2)

    points = []  # initialize a set of points
    # each next angle should step by a certain amount
    angle_stride = 2 * np.pi * phi_golden if geodesic else 2 * np.pi / phi_golden**2
    b = (
        n_b if n_b is not None else round(alpha * np.sqrt(n))
    )  # number of boundary points
    for k in range(1, n + 1):
        r = radius(k, n, b)  # get radius
        theta = k * angle_stride  # get angle
        points.append((r * np.cos(theta), r * np.sin(theta)))
    return points


class SunflowerFarmLayout(templates.LayoutTemplate):
    """
    A sunflower-inspired structured layout algorithm

    Options
    -------
    modeling_options : dict
        a modeling options dictionary (inherited from
        `templates.LayoutTemplate`)
    N_turbines : int
        the number of turbines that should be in the farm layout (inherited from
        `templates.LayoutTemplate`)

    Inputs
    ------
    alpha : float
        a parameter to control the number of boundary (v. interior) turbines
    spacing_target : float
        a parameter to control the target average minimum spacing

    Outputs
    -------
    x_turbines : np.ndarray
        a 1-D numpy array that represents that x (i.e. Easting) coordinate of
        the location of each of the turbines in the farm in meters (inherited
        from `templates.LayoutTemplate`)
    y_turbines : np.ndarray
        a 1-D numpy array that represents that y (i.e. Northing) coordinate of
        the location of each of the turbines in the farm in meters (inherited
        from `templates.LayoutTemplate`)
    spacing_effective_primary : float
        a measure of the spacing on a primary axis of a rectangular farm that
        would be equivalent to this one for the purposes of computing BOS costs
        measured in rotor diameters (inherited from `templates.LayoutTemplate`)
    spacing_effective_secondary : float
        a measure of the spacing on a secondary axis of a rectangular farm that
        would be equivalent to this one for the purposes of computing BOS costs
        measured in rotor diameters (inherited from `templates.LayoutTemplate`)
    """

    def initialize(self):
        """Initialization of OM component."""
        super().initialize()

    def setup(self):
        """Setup of OM component."""
        super().setup()

        # add parameters for sunflower farm DVs
        # self.add_input("alpha", 0.0, desc="boundary point control param.")
        self.add_input("spacing_target", 0.0, desc="target spacing in rotor diameters")

    def setup_partials(self):
        """Derivative setup for OM component."""

        # run FD for the layout tools
        self.declare_partials("*", "*", method="fd")

    def compute(self, inputs, outputs):
        """Computation for the OM component."""

        # get the desired mean nearest-neighbor distance
        D_rotor = self.windIO["wind_farm"]["turbine"][
            "rotor_diameter"
        ]  # get rotor diameter
        spacing_target = D_rotor * inputs["spacing_target"]

        # generate the points
        points = np.array(sunflower(self.N_turbines, geodesic=True))
        dist_mtx = scipy.spatial.distance.squareform(
            scipy.spatial.distance.pdist(points)
        )
        np.fill_diagonal(dist_mtx, np.inf)  # self-distance not meaningful, remove
        d_mean_NN = np.mean(np.min(dist_mtx, axis=0))

        # rescale points to achieve target spacing
        points *= spacing_target / d_mean_NN

        # pipe in the outputs
        outputs["x_turbines"] = points[:, 0].tolist()
        outputs["y_turbines"] = points[:, 1].tolist()
        outputs["spacing_effective_primary"] = spacing_target  # ???
        outputs["spacing_effective_secondary"] = spacing_target  # ???


class SunflowerFarmLanduse(fullfarm.FullFarmLanduse):
    pass  # just use the full-farm landuse component



import numpy as np

import openmdao.api as om


class CollectionTemplate(om.ExplicitComponent):
    """
    Template component for a energy collection system.

    A energy collection system component, based on this template, will compute the
    energy collection system necessary for a farm given its layout, turbine
    definitions, and substation location.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary

    Inputs
    ------
    x_turbines : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the turbines,
        with length `N_turbines`
    y_turbines : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the turbines,
        with length `N_turbines`
    x_substations : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the substations,
        with length `N_substations`
    y_substations : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the substations,
        with length `N_substations`

    Outputs
    -------
    total_length_cables : float
        the total length of cables used in the collection system network

    Discrete Outputs
    ----------------
    length_cables : np.ndarray
        a 1D numpy array that holds the lengths of each of the cables necessary
        to collect energy generated, with length `N_turbines`
    load_cables : np.ndarray
        a 1D numpy array that holds the turbine count upstream of the cable segment
        (i.e. number of turbines whose power is collected through the cable), with
        length `N_turbines`
    max_load_cables : int
        the maximum cable capacity required by the collection system
    terse_links : np.ndarray
        a 1D numpy int array encoding the electrical connections of the collection
        system (tree topology), with length `N_turbines`
    """

    def initialize(self):
        """Initialization of OM component."""
        self.options.declare("modeling_options")

    def setup(self):
        """Setup of OM component."""
        # load modeling options
        self.modeling_options = self.options["modeling_options"]
        self.windIO_plant = self.modeling_options["windIO_plant"]
        self.N_turbines = self.modeling_options["layout"]["N_turbines"]
        self.N_substations = self.modeling_options["layout"]["N_substations"]
        if "x_turbines" in self.modeling_options["layout"]:
            self.x_turbines = self.modeling_options["layout"]["x_turbines"]
        else:
            self.x_turbines = (
                self.windIO_plant.get("wind_farm", {})
                .get("layouts", {})
                .get("coordinates", {})
                .get("x", np.zeros((self.N_turbines,)))
            )
        if "y_turbines" in self.modeling_options["layout"]:
            self.y_turbines = self.modeling_options["layout"]["y_turbines"]
        else:
            self.y_turbines = (
                self.windIO_plant.get("wind_farm", {})
                .get("layouts", {})
                .get("coordinates", {})
                .get("y", np.zeros((self.N_turbines,)))
            )
        self.x_substations = np.array(
            [
                substation["electrical_substation"]["coordinates"]["x"]
                for substation in self.windIO_plant["wind_farm"][
                    "electrical_substations"
                ]
            ]
        )
        self.y_substations = np.array(
            [
                substation["electrical_substation"]["coordinates"]["y"]
                for substation in self.windIO_plant["wind_farm"][
                    "electrical_substations"
                ]
            ]
        )

        # set up inputs for farm layout
        self.add_input("x_turbines", self.x_turbines, units="m")
        self.add_input("y_turbines", self.y_turbines, units="m")
        self.add_input("x_substations", self.x_substations, units="m")
        self.add_input("y_substations", self.y_substations, units="m")
        self.add_discrete_input("x_border", None)
        self.add_discrete_input("y_border", None)

        # set up outputs for the collection system
        self.add_output("total_length_cables", 0.0, units="m")
        self.add_discrete_output("length_cables", np.zeros((self.N_turbines,)))
        self.add_discrete_output("terse_links", np.full((self.N_turbines,), -1))
        self.add_discrete_output("load_cables", np.zeros((self.N_turbines,)))
        self.add_discrete_output("max_load_cables", 0.0)
        self.add_discrete_output("graph", None)

    def compute(
        self,
        inputs,
        outputs,
        discrete_inputs=None,
        discrete_outputs=None,
    ):
        """
        Computation for the OM component.

        For a template class this is not implemented and raises an error!
        """

        ###########################################
        #                                         #
        # IMPLEMENT THE COLLECTION COMPONENT HERE #
        #                                         #
        ###########################################

        raise NotImplementedError(
            "This is an abstract class for a derived class to implement!"
        )



import numpy as np

import ard.farm_aero.templates as templates


class PlaceholderBatchPower(templates.BatchFarmPowerTemplate):
    """
    Placeholder component for computing power assuming nameplate power at all
    times.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary (inherited from `FarmAeroTemplate`)
    wind_query : floris.wind_data.WindRose
        a WindQuery objects that specifies the wind conditions that are to be
        computed

    Inputs
    ------
    x_turbines : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the turbines,
        with length `N_turbines` (inherited from `FarmAeroTemplate`)
    y_turbines : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the turbines,
        with length `N_turbines` (inherited from `FarmAeroTemplate`)
    yaw_turbines : np.ndarray
        a numpy array indicating the yaw angle to drive each turbine to with
        respect to the ambient wind direction, with length `N_turbines`
        (inherited from `FarmAeroTemplate`)

    Outputs
    -------
    power_farm : np.ndarray
        an array of the farm power for each of the wind conditions that have
        been queried
    power_turbines : np.ndarray
        an array of the farm power for each of the turbines in the farm across
        all of the conditions that have been queried on the wind rose
        (`N_turbines`, `N_wind_conditions`)
    thrust_turbines : np.ndarray
        an array of the wind turbine thrust for each of the turbines in the farm
        across all of the conditions that have been queried on the wind rose
        (`N_turbines`, `N_wind_conditions`)
    """

    def initialize(self):
        super().initialize()  # run super class script first!

    def setup(self):
        super().setup()

        # unpack wind query object
        self.wind_query = self.options["wind_query"]
        self.directions_wind = self.options["wind_query"].get_directions()
        self.speeds_wind = self.options["wind_query"].get_speeds()
        if self.options["wind_query"].get_TIs() is None:
            self.options["wind_query"].set_TI_using_IEC_method()
        self.TIs_wind = self.options["wind_query"].get_TIs()
        self.N_wind_conditions = self.options["wind_query"].N_conditions

        # add the outputs we want for a batched power analysis:
        #   - farm and turbine powers
        #   - turbine thrusts
        self.add_output(
            "power_farm",
            np.zeros((self.N_wind_conditions,)),
            units="W",
        )
        if self.return_turbine_output:
            self.add_output(
                "power_turbines",
                np.zeros((self.N_turbines, self.N_wind_conditions)),
                units="W",
            )
            self.add_output(
                "thrust_turbines",
                np.zeros((self.N_turbines, self.N_wind_conditions)),
                units="N",
            )

    def setup_partials(self):
        """Derivative setup for OM component."""
        # the default (but not preferred!) derivatives are FDM
        self.declare_partials("*", "*", method="exact")

    def compute(self, inputs, outputs):
        """
        Computation for the OM component.
        """

        # the following should be set
        outputs["power_farm"] = (
            self.N_turbines
            * self.modeling_options["turbine"]["nameplate"]["power_rated"]
            * np.ones((self.N_wind_conditions,))
        )
        outputs["power_turbines"] = self.modeling_options["turbine"]["nameplate"][
            "power_rated"
        ] * np.ones((self.N_turbines, self.N_wind_conditions))
        outputs["thrust_turbines"] = np.zeros((self.N_turbines, self.N_wind_conditions))


class PlaceholderAEP(templates.FarmAEPTemplate):
    """
    Placeholder component for computing AEP assuming nameplate power at all
    times.

    Options
    -------
    modeling_options : dict
        a modeling options dictionary (inherited via
        `templates.FarmAEPTemplate`)
    wind_query : floris.wind_data.WindRose
        a WindQuery objects that specifies the wind conditions that are to be
        computed (inherited from `templates.FarmAEPTemplate`)

    Inputs
    ------
    x_turbines : np.ndarray
        a 1D numpy array indicating the x-dimension locations of the turbines,
        with length `N_turbines` (inherited via `templates.FarmAEPTemplate`)
    y_turbines : np.ndarray
        a 1D numpy array indicating the y-dimension locations of the turbines,
        with length `N_turbines` (inherited via `templates.FarmAEPTemplate`)
    yaw_turbines : np.ndarray
        a numpy array indicating the yaw angle to drive each turbine to with
        respect to the ambient wind direction, with length `N_turbines`
        (inherited via `templates.FarmAEPTemplate`)

    Outputs
    -------
    AEP_farm : float
        the AEP of the farm given by the analysis (inherited from
        `templates.FarmAEPTemplate`)
    power_farm : np.ndarray
        an array of the farm power for each of the wind conditions that have
        been queried (inherited from `templates.FarmAEPTemplate`)
    power_turbines : np.ndarray
        an array of the farm power for each of the turbines in the farm across
        all of the conditions that have been queried on the wind rose
        (`N_turbines`, `N_wind_conditions`) (inherited from
        `templates.FarmAEPTemplate`)
    thrust_turbines : np.ndarray
        an array of the wind turbine thrust for each of the turbines in the farm
        across all of the conditions that have been queried on the wind rose
        (`N_turbines`, `N_wind_conditions`) (inherited from
        `templates.FarmAEPTemplate`)
    """

    def initialize(self):
        super().initialize()  # run super class script first!

    def setup(self):
        super().setup()  # run super class script first!

    def setup_partials(self):
        self.declare_partials("*", "*", method="exact")

    def compute(self, inputs, outputs):

        # the following should be set
        outputs["AEP_farm"] = (
            self.N_turbines
            * self.modeling_options["turbine"]["nameplate"]["power_rated"]
            * 1.0e6
            * 8760.0
        )
        outputs["power_farm"] = (
            self.N_turbines
            * self.modeling_options["turbine"]["nameplate"]["power_rated"]
            * 1.0e6
            * np.ones((self.N_wind_conditions,))
        )
        outputs["power_turbines"] = (
            self.modeling_options["turbine"]["nameplate"]["power_rated"]
            * 1.0e6
            * np.ones((self.N_turbines, self.N_wind_conditions))
        )
        outputs["thrust_turbines"] = np.zeros((self.N_turbines, self.N_wind_conditions))



import openmdao.api


def extract_pareto(
    prob: openmdao.api.Problem,
    ard_input_dict: dict,
):

    # grab the driver-scaled values off of the drivers
    desvar_nd = prob.driver.desvar_nd.copy()
    obj_nd = prob.driver.obj_nd.copy()

    #
